"""
Author: Ynag Wang
Date: 2023-11-02 14:52:08
LastEditTime: 2024-03-28 08:39:39
LastEditors: Wenyu Ouyang
Description: 
FilePath: \hydrodatasource\tests\test_sync.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os

import pytest

from hydrodatasource.configs import config
from hydrodatasource.reader.minio_api import minio_sync_files, boto3_sync_files

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
async def test_sync_data():
    s3_client = config.S3
    mc_client = config.MC
    await minio_sync_files(
        mc_client,
        "forestbat-private",
        local_path=os.path.join(config.LOCAL_DATA_PATH, "forestbat_test"),
    )
    await boto3_sync_files(
        s3_client,
        "forestbat-private",
        local_path=os.path.join(config.LOCAL_DATA_PATH, "forestbat_test_1"),
    )



"""
Author: Wenyu Ouyang
Date: 2023-10-31 09:26:31
LastEditTime: 2024-03-28 08:33:33
LastEditors: Wenyu Ouyang
Description: Interface for reader
FilePath: \hydrodatasource\hydrodatasource\reader\reader.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

from abc import ABC, abstractmethod
import os
import json
import datetime
import xarray as xr

from hydrodatasource.processor.gpm_gfs import make1nc41basin
from hydrodatasource.configs.config import LOCAL_DATA_PATH, GRID_INTERIM_BUCKET, RO, FS
from hydrodatasource.processor.mask import gen_single_mask


class AOI:
    def __init__(self, aoi_type, aoi_param):
        self._aoi_type = aoi_type  # can be "grid", "station", "basin" etc.
        self._aoi_param = aoi_param  # this can be a bounding box, coordinates, etc.

    @property
    def aoi_type(self):
        return self._aoi_type

    @property
    def aoi_param(self):
        return self._aoi_param

    def get_mask(self):
        # If it's a polygon, return its mask for data extraction
        if self._aoi_type == "basin":
            return gen_single_mask(
                basin_id=self._aoi_param,
                shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
                dataname="gpm",
                mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
            )
        else:
            # basin only for now
            raise NotImplementedError("Only basin available for now")

    # ... any other useful methods to describe or manipulate the AOI


class CommonHandler(AOI):
    def __init__(self, aoi_type, aoi_param, region=None, time_periods=None):
        super().__init__(aoi_type, aoi_param)
        self._region = region  # camels or wis, "camels" is on behalf of us, "wis" is on behalf of cn
        self._time_periods = time_periods

    @property
    def region(self):
        return self._region

    @property
    def time_periods(self):
        return self._time_periods

    def is_valid_time_periods(self):
        # 检查 time_periods 是否是列表
        if not isinstance(self._time_periods, list):
            return False

        for period in self._time_periods:
            # 检查每个元素是否是包含两个元素的列表
            if not isinstance(period, list) or len(period) != 2:
                return False

            # 检查每个元素是否是字符串且符合日期时间格式
            for date_str in period:
                if not isinstance(date_str, str):
                    return False
                try:
                    datetime.datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%S")
                except ValueError:
                    return False

        return True

    def read_file_from_minio(self):
        try:
            json_file_path = os.path.join(self._aoi_param, f"{self._dataname}.json")
            # 从 MinIO 读取 JSON 文件
            with FS.open(f"{GRID_INTERIM_BUCKET}/{json_file_path}") as f:
                json_data = json.load(f)
            # 使用 xarray 和 kerchunk 读取数据
            return xr.open_dataset(
                "reference://",
                engine="zarr",
                backend_kwargs={
                    "consolidated": False,
                    "storage_options": {
                        "fo": json_data,
                        "remote_protocol": "s3",
                        "remote_options": RO,
                    },
                },
            )
        except:
            raise FileNotFoundError(
                "Please check the file in the minio server. \
                This error is generally caused by the following two situations:\n\
                    1. There is no file for this basin in the minio server. Please make a new one and upload it.\n\
                    2. Check your settings. Make sure you have enough permission to access the minio server and the bucket."
            )


class DataHandler(CommonHandler):
    def __init__(
        self,
        aoi_type,
        aoi_param,
        region=None,
        time_periods=None,
        dataname=None,
        minio_read=True,
        local_save=True,
        minio_upload=False,
    ):
        super().__init__(aoi_type, aoi_param, region, time_periods)
        self._dataname = dataname
        self._minio_read = minio_read
        self._local_save = local_save
        self._minio_upload = minio_upload

    @property
    def dataname(self):
        return self._local_file_path

    @property
    def dataname(self):
        return self._dataname

    @property
    def minio_read(self):
        return self._minio_read

    @property
    def local_save(self):
        return self._local_save

    @property
    def minio_upload(self):
        return self._minio_upload

    def handle(self):
        if self._minio_read == True:
            return self.read_file_from_minio()
        else:
            return make1nc41basin(
                basin_id=self._aoi_param,
                dataname=self._dataname,
                local_path=LOCAL_DATA_PATH,
                mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
                shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
                dataset=self._region,
                time_periods=self._time_periods,
                local_save=self._local_save,
                minio_upload=self._minio_upload,
            )

    # TODO: time_periods is only used in make1nc41basin funtion, it needs to be used in other cases


class DataReaderStrategy(ABC):
    @abstractmethod
    def read(self, path: str, aoi: AOI):
        pass


class AbstractFileReader(DataReaderStrategy):
    def __init__(self, data_handler):
        self.data_handler = data_handler

    @abstractmethod
    def configure(self, path: str, aoi: AOI):
        pass

    def read(self, path: str, aoi: AOI):
        configuration = self.configure(path, aoi)
        return self.data_handler.handle(configuration)


class LocalFileReader(AbstractFileReader):
    def configure(self, path: str, aoi: AOI):
        return {"type": "local", "path": path, "aoi": aoi}


class MinioFileReader(AbstractFileReader):
    def __init__(self, minio_client, data_handler):
        super().__init__(data_handler)
        self.client = minio_client

    def configure(self, path: str, aoi: AOI):
        return {"type": "minio", "bucket": "your_bucket_name", "path": path, "aoi": aoi}



"""
Author: Yang Wang, Jingyi Wang, and Wenyu Ouyang
Date: 2023-10-28 09:23:22
LastEditTime: 2024-03-28 08:40:17
LastEditors: Wenyu Ouyang
Description: Test for rainfall-runoff event identification
FilePath: \hydrodatasource\tests\test_rr_event_iden.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import numpy as np
from pint import UnitRegistry

from hydrodataset import Camels

from hydrodatasource.configs.config import SETTING
from hydrodatasource.cleaner.dmca_esr import rainfall_runoff_event_identify
from hydrodatasource.utils.utils import streamflow_unit_conv


def test_rainfall_runoff_event_identify():
    camels = Camels(
        os.path.join(
            SETTING["local_data_path"]["datasets-origin"], "camels", "camels_us"
        )
    )
    gage_ids = camels.read_object_ids()
    ureg = UnitRegistry()

    rain = camels.read_ts_xrdataset(
        gage_ids[:1], ["1980-01-01", "2015-01-01"], var_lst=["prcp"]
    )
    flow = camels.read_ts_xrdataset(
        gage_ids[:1], ["1980-01-01", "2015-01-01"], var_lst=["streamflow"]
    )
    # trans unit to mm/day
    basin_area = camels.read_area(gage_ids[:1])
    r_mmd = streamflow_unit_conv(flow, basin_area)
    flow_threshold = streamflow_unit_conv(
        np.array([100]) * ureg.m**3 / ureg.s,
        basin_area.isel(basin=0).to_array().to_numpy() * ureg.km**2,
        target_unit="mm/h",
    )
    flood_events = rainfall_runoff_event_identify(
        rain["prcp"].isel(basin=0).to_series(),
        r_mmd["streamflow"].isel(basin=0).to_series(),
        flow_threshold=flow_threshold[0],
    )
    assert flood_events["BEGINNING_RAIN"].shape[0] == flood_events["END_RAIN"].shape[0]



from hydrodatasource.configs.config import SETTING
from datetime import datetime, timedelta
import pandas as pd
from loguru import logger
from sqlalchemy import create_engine


def read_forcing_dataframe(var_type, basin, time_period):
    start_time = time_period[0]
    end_time = time_period[1]
    if start_time is None:
        raise ValueError("the first element of the time_period cannot be None")

    table_name = {
        "gpm_tp": "t_gpm_pre_data",
        "gfs_tp": "t_gfs_tp_pre_data",
        "smap_sm_surface": "t_smap_pre_data",
        "gfs_soilw": "t_gfs_soil_pre_data",
    }

    column_dataname = {
        "gpm_tp": "tp",
        "smap_sm_surface": "sm_surface",
        "gfs_tp": "tp",
        "gfs_soilw": "sm_surface",
    }

    if var_type in ["gpm_tp", "smap_sm_surface"]:
        datetime_column = "predictdate"
    elif var_type in ["gfs_tp", "gfs_soilw"]:
        datetime_column = "forecastdatetime"

    if var_type not in table_name:
        raise ValueError(
            "var_type must be one of 'gpm_tp', 'gfs_tp', 'smap_sm_surface', 'gfs_soilw'"
        )
        
    # 将时间向前调整8小时
    start_time = datetime.strptime(start_time, '%Y-%m-%d %H:%M:%S')
    start_time -= timedelta(hours=8)
    if end_time:
        end_time = datetime.strptime(end_time, '%Y-%m-%d %H:%M:%S')
        end_time -= timedelta(hours=8)
        
    if var_type == "gpm_tp":
        sql = f"""
        SELECT 
            basincode, 
            predictdate, 
            data ->> 'tp' AS tp,
            data ->> 'raster_area' AS raster_area,
            data ->> 'intersection_area' AS intersection_area
        FROM (
            SELECT 
                basincode, 
                predictdate, 
                jsonb_array_elements(data) AS data
            FROM {table_name[var_type]}
        ) {table_name[var_type]}
        WHERE predictdate >= '{start_time}' 
        """
        if end_time is not None:
            sql += f" AND predictdate <= '{end_time}'"
        sql += f" AND basincode = '{basin}'"
    elif var_type == "smap_sm_surface":
        sql = f"""
        SELECT 
            basincode, 
            predictdate, 
            data ->> 'sm_surface' AS sm_surface,
            data ->> 'raster_area' AS raster_area,
            data ->> 'intersection_area' AS intersection_area
        FROM (
            SELECT 
                basincode, 
                predictdate, 
                jsonb_array_elements(data) AS data
            FROM {table_name[var_type]}
        ) {table_name[var_type]}
        WHERE predictdate >= '{start_time}' 
        """
        if end_time is not None:
            sql += f" AND predictdate <= '{end_time}'"
        sql += f" AND basincode = '{basin}'"
    elif var_type == "gfs_tp":
        sql = f"""
        select
            basin_code,
            forecastdatetime,
            tp AS tp,
            raster_area,
            intersection_area 
        from {table_name[var_type]}
        WHERE forecastdatetime >= '{start_time}' 
        """
        if end_time is not None:
            sql += f" AND forecastdatetime <= '{end_time}'"
        sql += f" AND basin_code = '{basin}'"
    elif var_type == "gfs_soilw":
        sql = f"""
        select
            basin_code,
            forecastdatetime,
            soilw AS sm_surface,
            raster_area,
            intersection_area 
        from {table_name[var_type]}
        WHERE forecastdatetime >= '{start_time}' 
        """
        if end_time is not None:
            sql += f" AND forecastdatetime <= '{end_time}'"
        sql += f" AND basin_code = '{basin}'"

    db_username = SETTING["postgres"]["username"]
    db_password = SETTING["postgres"]["password"]
    db_host = SETTING["postgres"]["server_url"]
    db_port = SETTING["postgres"]["port"]
    db_name = SETTING["postgres"]["database"]
    engine = create_engine(
        f"postgresql+psycopg2://{db_username}:{db_password}@{db_host}:{db_port}/{db_name}"
    )
    # 执行查询数据SQL查询
    try:
        df = pd.read_sql(sql, engine)
        # 转换数据类型
        df[column_dataname[var_type]] = df[column_dataname[var_type]].astype(float)
        df["raster_area"] = df["raster_area"].astype(float)
        df["intersection_area"] = df["intersection_area"].astype(float)
        # 按照时间列排序
        df = df.sort_values(by=datetime_column)
        
        df[datetime_column] = pd.to_datetime(df[datetime_column]) + timedelta(hours=8)
    except Exception as e:
        logger.error(e)
        raise Exception() from e

    return df

def read_plcd(basin):
    sql = """
    SELECT
        stcd,
        plcd
    FROM
        t_xaj_parameter
    WHERE
        stcd = %s
    """
    
    db_username = SETTING["postgres"]["username"]
    db_password = SETTING["postgres"]["password"]
    db_host = SETTING["postgres"]["server_url"]
    db_port = SETTING["postgres"]["port"]
    db_name = SETTING["postgres"]["database"]
    
    engine = create_engine(
        f"postgresql+psycopg2://{db_username}:{db_password}@{db_host}:{db_port}/{db_name}"
    )
    
    result = pd.read_sql(sql, engine, params=(basin,))
    
    if result.empty:
        return None
    
    return result.iloc[0]['plcd']


import collections
import json
import os
import re
import numpy as np
import pandas as pd
import xarray as xr
from abc import ABC
from tqdm import tqdm

from hydroutils import hydro_file
from hydroutils.hydro_time import generate_start0101_time_range

import hydrodatasource.configs.config as conf
from hydrodatasource.configs.data_consts import (
    ERA5LAND_ET_REALATED_VARS,
    MODIS_ET_PET_8D_VARS,
)
from hydrodatasource.utils.utils import (
    calculate_basin_offsets,
    is_minio_folder,
    minio_file_list,
)
from hydrodatasource.reader import access_fs

CACHE_DIR = hydro_file.get_cache_dir()


class HydroData(ABC):
    """An interface for reading multi-modal data sources.

    Parameters
    ----------
    ABC : _type_
        _description_
    """

    def __init__(self, data_path):
        self.data_source_dir = data_path

    def get_name(self):
        raise NotImplementedError

    def set_data_source_describe(self):
        raise NotImplementedError

    def read_data(self):
        raise NotImplementedError


class SelfMadeHydroDataset(HydroData):
    """A class for reading hydrodataset, but not really ready-datasets,
    just some data directorys organized like a HydroDataset.

    NOTE:
    We compile forcing data and attr data into a directory,
    organized like a ready dataset -- like Caravan.
    Only two directories are needed: attributes and timeseries
    """

    def __init__(self, data_path, download=False, time_unit=None):
        """Initialize a self-made Caravan-style dataset.

        Parameters
        ----------
        data_path : _type_
            _description_
        download : bool, optional
            _description_, by default False
        time_unit : list, optional
            _description_, by default
        """
        if time_unit is None:
            time_unit = ["1D"]
        if any(unit not in ["1h", "3h", "1D", "8D"] for unit in time_unit):
            raise ValueError(
                "time_unit must be one of ['1h', '3h', '1D', '8D']. We only support these time units now."
            )
        # TODO: maybe starting with "s3://" is a better idea?
        self.head = "minio" if "s3://" in data_path else "local"
        super().__init__(data_path)
        self.data_source_description = self.set_data_source_describe()
        if download:
            self.download_data_source()
        self.camels_sites = self.read_site_info()
        self.time_unit = time_unit

    @property
    def streamflow_unit(self):
        unit_mapping = {"1h": "mm/h", "3h": "mm/3h", "1D": "mm/d"}
        return {unit: unit_mapping[unit] for unit in self.time_unit}

    def get_name(self):
        return "SelfMadeHydroDataset"

    def set_data_source_describe(self):
        data_root_dir = self.data_source_dir
        ts_dir = os.path.join(data_root_dir, "timeseries")
        # we assume that each subdirectory in ts_dir represents a time unit
        # In this subdirectory, there are csv files for each basin
        if "s3://" in data_root_dir:
            time_units_dir = [
                os.path.join(ts_dir, name)
                for name in minio_file_list(ts_dir)
                if is_minio_folder(os.path.join(ts_dir, name))
            ]
        else:
            time_units_dir = [
                os.path.join(ts_dir, name)
                for name in os.listdir(ts_dir)
                if os.path.isdir(os.path.join(ts_dir, name))
            ]
        unit_files = [folder + "_units_info.json" for folder in time_units_dir]
        attr_dir = os.path.join(data_root_dir, "attributes")
        attr_file = os.path.join(attr_dir, "attributes.csv")
        shape_dir = os.path.join(data_root_dir, "shapes")

        return collections.OrderedDict(
            DATA_DIR=data_root_dir,
            TS_DIRS=time_units_dir,
            ATTR_DIR=attr_dir,
            ATTR_FILE=attr_file,
            UNIT_FILES=unit_files,
            SHAPE_DIR=shape_dir,
        )

    def download_data_source(self):
        print(
            "Please download it manually and put all files of a CAMELS dataset in the CAMELS_DIR directory."
        )
        print("We unzip all files now.")

    def read_site_info(self):
        camels_file = self.data_source_description["ATTR_FILE"]
        attrs = access_fs.spec_path(camels_file, head=self.head)
        return attrs[["basin_id", "area"]]

    def read_object_ids(self, object_params=None) -> np.array:
        return self.camels_sites["basin_id"].values

    def read_timeseries(
        self, object_ids=None, t_range_list: list = None, relevant_cols=None, **kwargs
    ) -> dict:
        """
        Returns a dictionary containing data with different time scales.

        Parameters
        ----------
        object_ids : list, optional
            List of object IDs. Defaults to None.
        t_range_list : list, optional
            List of time ranges. Defaults to None.
        relevant_cols : list, optional
            List of relevant columns. Defaults to None.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        dict
            A dictionary containing data with different time scales.
        """
        time_units = kwargs.get("time_units", ["1D"])
        region = kwargs.get("region", None)
        start0101_freq = kwargs.get("start0101_freq", False)

        results = {}

        for time_unit in time_units:
            # whether to convert the time to UTC, for 1D time unit, default set False,
            # and for 3h time unit, set True
            offset_to_utc = time_unit == "3h"
            if offset_to_utc:
                basinoutlets_path = os.path.join(
                    self.data_source_description["SHAPE_DIR"], "basinoutlets.shp"
                )
                try:
                    offset_dict = calculate_basin_offsets(basinoutlets_path)
                except:
                    raise FileNotFoundError(
                        f"basinoutlets.shp not found in {basinoutlets_path}."
                    )
            ts_dir = next(
                dir_path
                for dir_path in self.data_source_description["TS_DIRS"]
                if time_unit in dir_path
            )
            if start0101_freq:
                t_range = generate_start0101_time_range(
                    start_time=t_range_list[0],
                    end_time=t_range_list[-1],
                    freq=time_unit,
                )
            else:
                t_range = pd.date_range(
                    start=t_range_list[0], end=t_range_list[-1], freq=time_unit
                )
            nt = len(t_range)
            x = np.full([len(object_ids), nt, len(relevant_cols)], np.nan)

            for k in tqdm(
                range(len(object_ids)), desc=f"Reading timeseries data with {time_unit}"
            ):
                prefix_ = "" if region is None else region + "_"
                ts_file = os.path.join(
                    ts_dir,
                    prefix_ + object_ids[k] + ".csv",
                )
                if "s3://" in ts_file:
                    with conf.FS.open(ts_file, mode="rb") as f:
                        ts_data = pd.read_csv(f)
                else:
                    ts_data = pd.read_csv(ts_file)
                date = pd.to_datetime(ts_data["time"]).values
                if offset_to_utc:
                    date = date - np.timedelta64(offset_dict[object_ids[k]], "h")
                [_, ind1, ind2] = np.intersect1d(date, t_range, return_indices=True)

                for j in range(len(relevant_cols)):
                    tmp_ = self._read_timeseries_1basin1var(ts_data, relevant_cols[j])
                    x[k, ind2, j] = tmp_[ind1]

            results[time_unit] = x

        return results

    def _read_timeseries_1basin1var(self, ts_data, relevant_col):
        if "precipitation" in relevant_col:
            prcp = ts_data[relevant_col].values
            prcp[prcp < 0] = 0.0
            return prcp
        elif relevant_col in ERA5LAND_ET_REALATED_VARS:
            evap = -1 * ts_data[relevant_col].values
            evap[evap < 0] = 0.0
            return evap
        elif relevant_col in MODIS_ET_PET_8D_VARS:
            modis_values = ts_data[relevant_col].values
            modis_dates = pd.to_datetime(ts_data["time"].values)
            for idx, current_date in enumerate(modis_dates):
                # Check if the date is prior to or on January 1st
                if current_date.month == 1 and current_date.day == 1:
                    if idx == 0:
                        # First day is January 1st, no previous date to scale from
                        continue
                        # Get the previous date
                    previous_date = modis_dates[idx - 1]
                    # Calculate the number of days between the previous date and January 1st
                    delta_days = (current_date - previous_date).days

                    # Adjust the MODIS value based on the number of days between the previous date and January 1st
                    if delta_days > 0:
                        modis_values[idx - 1] = modis_values[idx - 1] * 8 / delta_days
            # NOTE: MODIS ET values are ACTUALLY in 0.1mm/day, so we need to convert to mm/day
            return modis_values * 0.1
        else:
            return ts_data[relevant_col].values

    def read_attributes(
        self, object_ids=None, constant_cols=None, **kwargs
    ) -> np.array:
        """2d data (site_num * var_num), non-time-series data"""
        attr_file = self.data_source_description["ATTR_FILE"]
        if "s3://" in attr_file:
            with conf.FS.open(attr_file, mode="rb") as f:
                attrs = pd.read_csv(f, dtype={"basin_id": str})
        else:
            attrs = pd.read_csv(attr_file, dtype={"basin_id": str})
        if object_ids is None:
            if constant_cols is None:
                return attrs
            object_ids = attrs["basin_id"].values
        if constant_cols is None:
            constant_cols = attrs.columns.values
        x = np.full([len(object_ids), len(constant_cols)], np.nan)
        for k in range(len(object_ids)):
            ind = attrs["basin_id"] == object_ids[k]
            for j in range(len(constant_cols)):
                x[k, j] = attrs[constant_cols[j]][ind].values
        return x

    def get_attributes_cols(self) -> np.array:
        """the constant cols in this data_source"""
        attr_file = self.data_source_description["ATTR_FILE"]
        if "s3://" in attr_file:
            with conf.FS.open(attr_file, mode="rb") as f:
                attrs = pd.read_csv(f, dtype={"basin_id": str})
        else:
            attrs = pd.read_csv(attr_file, dtype={"basin_id": str})
        attr_units = attrs.columns.values
        return self._check_vars_in_unitsinfo(attr_units)

    def get_timeseries_cols(self) -> np.array:
        """the relevant cols in this data_source"""
        ts_dirs = self.data_source_description["TS_DIRS"]
        unit_files = self.data_source_description["UNIT_FILES"]
        all_vars = {}
        for time_unit in self.time_unit:
            # Find the directory that corresponds to the current time unit
            ts_dir = next(dir_path for dir_path in ts_dirs if time_unit in dir_path)
            # Find the corresponding unit file
            unit_file = next(file for file in unit_files if time_unit in file)
            # Load the first CSV file in the directory to extract column names
            if "s3://" in ts_dir:
                ts_file = os.path.join(ts_dir, minio_file_list(ts_dir)[0])
                with conf.FS.open(ts_file, mode="rb") as f:
                    ts_tmp = pd.read_csv(f, dtype={"basin_id": str})
            else:
                ts_file = os.path.join(ts_dir, os.listdir(ts_dir)[0])
                ts_tmp = pd.read_csv(ts_file, dtype={"basin_id": str})
            # Get the relevant forcing units and validate against unit info
            forcing_units = ts_tmp.columns.values[1:]
            the_vars = self._check_vars_in_unitsinfo(forcing_units, unit_file)
            # Map the variables to the corresponding time unit
            all_vars[time_unit] = the_vars
        return all_vars

    def _check_vars_in_unitsinfo(self, vars, unit_file=None):
        """If a var is not recorded in a units_info file, we will not use it.

        Parameters
        ----------
        vars : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        if unit_file is None:
            # For attributes, all the variables' units are same in all unit_info files
            # hence, we just chose the first one
            unit_file = self.data_source_description["UNIT_FILES"][0]
        if "s3://" in unit_file:
            with conf.FS.open(unit_file, mode="rb") as fp:
                units_info = json.load(fp)
        else:
            units_info = hydro_file.unserialize_json(unit_file)
        vars_final = [var_ for var_ in vars if var_ in units_info]
        return np.array(vars_final)

    def cache_attributes_xrdataset(self, region=None):
        """Convert all the attributes to a single dataset

        Returns
        -------
        None
        """
        # NOTICE: although it seems that we don't use pint_xarray, we have to import this package
        import pint_xarray  # noqa: F401

        df_attr = self.read_attributes()
        df_attr.set_index("basin_id", inplace=True)
        # Mapping provided units to the variables in the datasets
        # For attributes, all the variables' units are same in all unit_info files
        # hence, we just chose the first one
        if "s3://" in self.data_source_description["UNIT_FILES"][0]:
            with conf.FS.open(
                self.data_source_description["UNIT_FILES"][0], mode="rb"
            ) as fp:
                units_dict = json.load(fp)
        else:
            units_dict = hydro_file.unserialize_json(
                self.data_source_description["UNIT_FILES"][0]
            )

        # Convert string columns to categorical variables and record categorical mappings
        categorical_mappings = {}
        for column in df_attr.columns:
            if df_attr[column].dtype == "object":
                df_attr[column] = df_attr[column].astype("category")
                categorical_mappings[column] = dict(
                    enumerate(df_attr[column].cat.categories)
                )
                df_attr[column] = df_attr[column].cat.codes

        ds = xr.Dataset()
        for column in df_attr.columns:
            attrs = {"units": units_dict.get(column, "unknown")}
            if column in categorical_mappings:
                attrs["category_mapping"] = categorical_mappings[column]

            data_array = xr.DataArray(
                data=df_attr[column].values,
                dims=["basin"],
                # we have set gage_id as index so that it won't be saved as numeric values
                coords={"basin": df_attr.index.values.astype(str)},
                attrs=attrs,
            )
            ds[column] = data_array

        # Convert categorical mappings to strings
        for column in ds.data_vars:
            if "category_mapping" in ds[column].attrs:
                # Convert the dictionary to a string
                mapping_str = str(ds[column].attrs["category_mapping"])
                ds[column].attrs["category_mapping"] = mapping_str
        prefix_ = "" if region is None else region + "_"
        ds.to_netcdf(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))

    def cache_timeseries_xrdataset(self, region=None, t_range=None, **kwargs):
        """Save all timeseries data in separate NetCDF files for each time unit.

        Parameters
        ----------
        region : str, optional
            A prefix used in cache file, by default None
        t_range : list, optional
            Time range for the data, by default ["1980-01-01", "2023-12-31"]
        kwargs : dict, optional
            batchsize -- Number of basins to process per batch, by default 100
            time_units -- List of time units to process, by default None
            start0101_freq -- for freq setting, if the start date is 01-01, set True, by default False
        """
        batchsize = kwargs.get("batchsize", 100)
        time_units = kwargs.get("time_units", self.time_unit) or [
            "1D"
        ]  # Default to ["1D"] if not specified or if time_units is None
        start0101_freq = kwargs.get("start0101_freq", False)

        variables = self.get_timeseries_cols()
        basins = self.camels_sites["basin_id"].values

        # Define the generator function for batching
        def data_generator(basins, batch_size):
            for i in range(0, len(basins), batch_size):
                yield basins[i : i + batch_size]

        for time_unit in time_units:
            if t_range is None:
                if time_unit != "3h":
                    t_range = ["1980-01-01", "2023-12-31"]
                else:
                    t_range = ["1980-01-01 01", "2023-12-31 22"]

            # Generate the time range specific to the time unit
            if start0101_freq:
                times = (
                    generate_start0101_time_range(
                        start_time=t_range[0], end_time=t_range[-1], freq=time_unit
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            else:
                times = (
                    pd.date_range(start=t_range[0], end=t_range[-1], freq=time_unit)
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            # Retrieve the correct units information for this time unit
            unit_file = next(
                file
                for file in self.data_source_description["UNIT_FILES"]
                if time_unit in file
            )
            if "s3://" in unit_file:
                with conf.FS.open(unit_file, mode="rb") as fp:
                    units_info = json.load(fp)
            else:
                units_info = hydro_file.unserialize_json(unit_file)

            for basin_batch in data_generator(basins, batchsize):
                data = self.read_timeseries(
                    object_ids=basin_batch,
                    t_range_list=t_range,
                    relevant_cols=variables[
                        time_unit
                    ],  # Ensure we use the right columns for the time unit
                    time_units=[
                        time_unit
                    ],  # Pass the time unit to ensure correct data retrieval
                    start0101_freq=start0101_freq,
                )

                dataset = xr.Dataset(
                    data_vars={
                        variables[time_unit][i]: (
                            ["basin", "time"],
                            data[time_unit][:, :, i],
                            {"units": units_info[variables[time_unit][i]]},
                        )
                        for i in range(len(variables[time_unit]))
                    },
                    coords={
                        "basin": basin_batch,
                        "time": pd.to_datetime(times),
                    },
                )

                # Save the dataset to a NetCDF file for the current batch and time unit
                prefix_ = "" if region is None else region + "_"
                batch_file_path = os.path.join(
                    CACHE_DIR,
                    f"{prefix_}timeseries_{time_unit}_batch_{basin_batch[0]}_{basin_batch[-1]}.nc",
                )
                dataset.to_netcdf(batch_file_path)

                # Release memory by deleting the dataset
                del dataset
                del data

    def cache_xrdataset(self, region=None, t_range=None, time_units=None):
        """Save all data in a netcdf file in the cache directory"""
        self.cache_attributes_xrdataset(region=region)
        self.cache_timeseries_xrdataset(
            region=region, t_range=t_range, time_units=time_units
        )

    def read_ts_xrdataset(
        self,
        gage_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ) -> dict:
        """
        Read time-series xarray dataset from multiple NetCDF files and organize them by time units.

        Parameters:
        ----------
        gage_id_lst: list - List of gage IDs to select.
        t_range: list - List of two elements [start_time, end_time] to select time range.
        var_lst: list - List of variables to select.
        **kwargs: Additional arguments.

        Returns:
        ----------
        dict: A dictionary where each key is a time unit and each value is an xarray.Dataset containing the selected gage IDs, time range, and variables.
        """
        region = kwargs.get("region", None)
        time_units = kwargs.get("time_units", self.time_unit)
        if var_lst is None:
            return None

        # Initialize a dictionary to hold datasets for each time unit
        datasets_by_time_unit = {}

        prefix_ = "" if region is None else region + "_"

        for time_unit in time_units:
            # Collect batch files specific to the current time unit
            batch_files = [
                os.path.join(CACHE_DIR, f)
                for f in os.listdir(CACHE_DIR)
                if re.match(
                    rf"^{prefix_}timeseries_{time_unit}_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                    f,
                )
            ]

            if not batch_files:
                # Cache the data if no batch files are found for the current time unit
                self.cache_timeseries_xrdataset(region=region, **kwargs)
                batch_files = [
                    os.path.join(CACHE_DIR, f)
                    for f in os.listdir(CACHE_DIR)
                    if re.match(
                        rf"^{prefix_}timeseries_{time_unit}_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                        f,
                    )
                ]

            selected_datasets = []

            for batch_file in batch_files:
                ds = xr.open_dataset(batch_file)
                all_vars = ds.data_vars
                if any(var not in ds.variables for var in var_lst):
                    raise ValueError(f"var_lst must all be in {all_vars}")
                if valid_gage_ids := [
                    gid for gid in gage_id_lst if gid in ds["basin"].values
                ]:
                    ds_selected = ds[var_lst].sel(
                        basin=valid_gage_ids, time=slice(t_range[0], t_range[1])
                    )
                    selected_datasets.append(ds_selected)

                ds.close()  # Close the dataset to free memory

            # If any datasets were selected, concatenate them along the 'basin' dimension
            if selected_datasets:
                datasets_by_time_unit[time_unit] = xr.concat(
                    selected_datasets, dim="basin"
                )
            else:
                datasets_by_time_unit[time_unit] = xr.Dataset()

        return datasets_by_time_unit

    def read_attr_xrdataset(self, gage_id_lst=None, var_lst=None, **kwargs):
        region = kwargs.get("region", None)

        prefix_ = "" if region is None else region + "_"
        if var_lst is None or len(var_lst) == 0:
            return None
        try:
            attr = xr.open_dataset(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
        except FileNotFoundError:
            self.cache_xrdataset(time_units=self.time_unit)
            attr = xr.open_dataset(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
        return attr[var_lst].sel(basin=gage_id_lst)

    def read_area(self, gage_id_lst=None):
        """read area of each basin/unit"""
        return self.read_attr_xrdataset(gage_id_lst, ["area"])

    def read_mean_prcp(self, gage_id_lst=None):
        """read mean precipitation of each basin/unit"""
        return self.read_attr_xrdataset(gage_id_lst, ["pre_mm_syr"])



"""
Author: Wenyu Ouyang
Date: 2024-03-24 08:48:57
LastEditTime: 2024-03-28 08:38:13
LastEditors: Wenyu Ouyang
Description: Check user's data format is correct (local and minio)
FilePath: \hydrodatasource\hydrodatasource\processor\data_checker.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pandas as pd
from hydrodatasource.configs.config import (
    LOCAL_DATA_PATH,
)


class DataChecker:
    """
    A class to check the format of station, reservoir, and basin data based on specified rules.
    """

    def __init__(self):
        """
        Initialize the DataChecker with the path to the data.
        """
        self.base_path = LOCAL_DATA_PATH
        # TODO: all specifications should be checked again
        # NOTE: we prioritize the basin data format check as it is directly related to the model
        self.expected_structure = {
            "stations-origin": [
                "pp_stations",
                "zq_stations",
                "zz_stations",
                "stations_list",
            ],
            "reservoirs-origin": ["day_data", "hour_data", "reservoirs_list"],
            "basins-origin": [
                "basins_list",
                "basin_shapefiles",
                "hour_data",
                "attributes.nc",
                "basins_shp.zip",
                "HydroRIVERS_v10_shp.zip",
            ],
        }
        self.expected_columns = {
            # 站点基础信息表
            "stations_list": [
                "ID",
                "STCD",
                "STTYPE",
                "VARTYPE",
                "DIVISION",
                "LON",
                "LAT",
            ],
            # 雨量站时序数据表
            "pp_stations": ["ID", "STCD", "TM", "DRP", "INTV", "PDR", "DYP", "WTH"],
            # 河道水文站/水位站时序信息表
            "zq_stations": [
                "ID",
                "STCD",
                "TM",
                "Z",
                "Q",
                "XSA",
                "XSAVV",
                "XSMXV",
                "FLWCHRCD",
                "WPTN",
                "MSQMT",
                "MSAMT",
                "MSVMT",
            ],
            "zz_stations": [
                "ID",
                "STCD",
                "TM",
                "Z",
                "Q",
                "XSA",
                "XSAVV",
                "XSMXV",
                "FLWCHRCD",
                "WPTN",
                "MSQMT",
                "MSAMT",
                "MSVMT",
            ],
            # 水库基础信息表
            "reservoirs_list": [
                "ID",
                "STCD",
                "TTCP",
                "DDZ",
                "NORMZ",
                "FSLTDZ",
                "LON",
                "LAT",
                "BASIN",
            ],
            # 水库时序数据表
            "reservoirs": [
                "ID",
                "STCD",
                "TM",
                "RZ",
                "INQ",
                "W",
                "BLRZ",
                "OTQ",
                "RWCHRCD",
                "RWPTN",
                "INQDR",
                "MSQMT",
            ],
            # 流域属性数据
            "basins_attributes": [
                "basin",
                "area",
                "ele_mt_smn",
                "slp_dg_sav",
                "sgr_dk_sav",
                "for_pc_sse",
                "glc_cl_smj",
                "run_mm_syr",
                "inu_pc_slt",
                "cmi_ix_syr",
                "aet_mm_syr",
                "snw_pc_syr",
                "swc_pc_syr",
                "gwt_cm_sav",
                "cly_pc_sav",
                "dor_pc_pva",
            ],
            # 流域时序平均数据表
            "basins_mean_data": [
                # 根据文档提供的信息定义列名，例如:
                "BAS_CODE",
                "TS",
                "GPM_TP",
                "GFS_TP",
                "GFS_10U",
                "GFS_10V",
                "GFS_2SH",
                "GFS_2R",
                "GFS_DSWRF",
                "GFS_PWAT",
                "GFS_2T",
                "GFS_TCC",
                "SMP",
                "Q",
                # 注意：这里的列名需要根据文档中的实际定义进行调整
            ],
            # 流域时序网格数据
            "basins_grid_data": [
                # 根据文档中的具体格式定义列名
                # 例如:
                "BAS_CODE",
                "TS",
                "GRID_DATA_TYPE",
                "GRID_VALUE",
                # 注意：这里的列名需要根据文档中的实际定义进行调整
            ],
            # others should be check again and more need to be added
        }

    def check_folder_structure(self):
        """
        Checks if the folder structure in the base directory matches the expected structure.

        Returns
        -------
        bool
            True if the structure matches, False otherwise.
        """
        for main_folder, subfolders in self.expected_structure.items():
            main_folder_path = os.path.join(self.base_path, main_folder)
            if not os.path.exists(main_folder_path):
                print(f"Missing main folder: {main_folder}")
                return False
            for subfolder in subfolders:
                subfolder_path = os.path.join(main_folder_path, subfolder)
                if not os.path.exists(subfolder_path):
                    print(f"Missing subfolder: {subfolder} in {main_folder}")
                    return False
        return True

    def check_station_data_files(self, station_types=None):
        """
        Checks the format of station data files including both basic info and time-series data.
        """
        if station_types is None:
            # one can specify the station types to check,
            # such as ["pp"], ["zz"], or ["zq"]
            station_types = ["pp", "zz", "zq"]
        is_correct = True
        for station_type in station_types:
            # Check basic info file
            basic_info_file = os.path.join(
                # here we use / rather than os.path.join to avoid path issues for minio
                self.base_path,
                f"stations-origin/{station_type}_stations.csv",
            )
            if not self.check_file_format(
                basic_info_file, self.expected_columns[f"{station_type}_stations"]
            ):
                print(f"Basic info file format error: {basic_info_file}")
                is_correct = False

            # Check time-series data files
            time_series_folder = os.path.join(
                self.base_path, f"stations-origin/{station_type}_stations"
            )
            if not self.check_files_in_folder(
                time_series_folder, self.expected_columns[f"{station_type}_stations"]
            ):
                print(f"Time series file format error in folder: {time_series_folder}")
                is_correct = False

        return is_correct

    def check_file_format(self, file_path, expected_columns):
        """
        Checks the format of a given file.

        Parameters
        ----------
        file_path : str
            Path to the file.
        expected_columns : list
            List of expected column names.

        Returns
        -------
        bool
            True if the file format is correct, False otherwise.
        """
        try:
            data = pd.read_csv(file_path)
            return all(column in data.columns for column in expected_columns)
        except Exception as e:
            print(f"Error reading file {file_path}: {e}")
            return False

    def check_files_in_folder(self, folder_path, expected_columns):
        """
        Checks the format of all CSV files in a folder.

        Parameters
        ----------
        folder_path : str
            Path to the folder containing the files.
        expected_columns : list
            List of expected column names for the files.

        Returns
        -------
        bool
            True if all file formats are correct, False otherwise.
        """
        is_correct = True
        for file in os.listdir(folder_path):
            if file.endswith(".csv"):
                file_path = os.path.join(folder_path, file)
                if not self.check_file_format(file_path, expected_columns):
                    print(f"File format error: {file_path}")
                    is_correct = False
        return is_correct

    def check_basin_data_files(self):
        """
        Checks the format of basin data files including attributes and time-series data.
        """
        is_correct = True

        # Check basin attributes data file
        basin_attributes_file = os.path.join(
            self.base_path, "basins-origin/attributes.nc"
        )
        if not os.path.isfile(basin_attributes_file):
            print(f"Missing basin attributes data file: {basin_attributes_file}")
            is_correct = False

        # Check basin time-series data file
        basin_time_series_file = os.path.join(
            self.base_path, "basins-origin/basins_mean_data.csv"
        )
        if not self.check_file_format(
            basin_time_series_file, self.expected_columns["basins_mean_data"]
        ):
            print(f"Basin time series data file format error: {basin_time_series_file}")
            is_correct = False

        # Check for spatial data files, if required
        # Example: checking for a specific shapefile
        basin_shapefile = os.path.join(self.base_path, "basins-origin/basins_shp.zip")
        if not os.path.isfile(basin_shapefile):
            print(f"Missing basin shapefile: {basin_shapefile}")
            is_correct = False

        return is_correct

    def check_basin_average_time_series_data(self):
        """
        Checks the format of basin average time-series data file.
        """
        # Define the file path for basin average time-series data
        basin_avg_time_series_file = os.path.join(
            self.base_path, "basins-origin/basin_average_time_series.csv"
        )

        # Check if the file exists
        if not os.path.isfile(basin_avg_time_series_file):
            print(
                f"Missing basin average time series data file: {basin_avg_time_series_file}"
            )
            return False

        # Check the file format
        if not self.check_file_format(
            basin_avg_time_series_file, self.expected_columns["basins_mean_data"]
        ):
            print(
                f"Basin average time series data file format error: {basin_avg_time_series_file}"
            )
            return False

        return True



"""
该模块用于从minio中读取era5-land数据，主要方法包括：

- `open_dataset` - 普通读取数据方法
- `from_shp` - 通过已有矢量范围读取数据方法
- `from_aoi` - 通过已有GeoDataFrame范围读取数据方法
- `to_netcdf` - 保存到本地文件
"""

from ..configs.config import FS, MINIO_PARAM
from ..configs.config import RO
from ..utils.utils import regen_box
import os
import s3fs
import numpy as np
import xarray as xr

bucket_name = MINIO_PARAM["bucket_name"]

start = np.datetime64("2015-01-01T00:00:00.000000000")
end = np.datetime64("2021-12-31T23:00:00.000000000")
box = (115, 38, 136, 54)
variables = [
    "10 metre U wind component",
    "10 metre V wind component",
    "2 metre dewpoint temperature",
    "2 metre temperature",
    "Evaporation",
    "Evaporation from bare soil",
    "Evaporation from open water surfaces excluding oceans",
    "Evaporation from the top of canopy",
    "Evaporation from vegetation transpiration",
    "Forecast albedo",
    "Lake bottom temperature",
    "Lake ice total depth",
    "Lake ice surface temperature",
    "Lake mix-layer depth",
    "Lake mix-layer temperature",
    "Lake shape factor",
    "Lake total layer temperature",
    "Leaf area index, high vegetation",
    "Leaf area index, low vegetation",
    "Potential evaporation",
    "Runoff",
    "Skin reservoir content",
    "Skin temperature",
    "Snow albedo",
    "Snow cover",
    "Snow density",
    "Snow depth",
    "Snow depth water equivalent",
    "Snow evaporation",
    "Snowfall",
    "Snowmelt",
    "Soil temperature level 1",
    "Soil temperature level 2",
    "Soil temperature level 3",
    "Soil temperature level 4",
    "Sub-surface runoff",
    "Surface latent heat flux",
    "Surface net solar radiation",
    "Surface net thermal radiation",
    "Surface pressure",
    "Surface runoff",
    "Surface sensible heat flux",
    "Surface solar radiation downwards",
    "Surface thermal radiation downwards",
    "Temperature of snow layer",
    "Total precipitation",
    "Volumetric soil water layer 1",
    "Volumetric soil water layer 2",
    "Volumetric soil water layer 3",
    "Volumetric soil water layer 4",
]
accumulated = [
    # '10 metre U wind component',
    # '10 metre V wind component',
    # '2 metre dewpoint temperature',
    # '2 metre temperature',
    "Evaporation",
    "Evaporation from bare soil",
    "Evaporation from open water surfaces excluding oceans",
    "Evaporation from the top of canopy",
    "Evaporation from vegetation transpiration",
    # 'Forecast albedo',
    # 'Lake bottom temperature',
    # 'Lake ice total depth',
    # 'Lake ice surface temperature',
    # 'Lake mix-layer depth',
    # 'Lake mix-layer temperature',
    # 'Lake shape factor',
    # 'Lake total layer temperature',
    # 'Leaf area index, high vegetation',
    # 'Leaf area index, low vegetation',
    "Potential evaporation",
    "Runoff",
    # 'Skin reservoir content',
    # 'Skin temperature',
    # 'Snow albedo',
    # 'Snow cover',
    # 'Snow density',
    # 'Snow depth',
    # 'Snow depth water equivalent',
    "Snow evaporation",
    "Snowfall",
    "Snowmelt",
    # 'Soil temperature level 1',
    # 'Soil temperature level 2',
    # 'Soil temperature level 3',
    # 'Soil temperature level 4',
    "Sub-surface runoff",
    "Surface latent heat flux",
    "Surface net solar radiation",
    "Surface net thermal radiation",
    # 'Surface pressure',
    "Surface runoff",
    "Surface sensible heat flux",
    "Surface solar radiation downwards",
    "Surface thermal radiation downwards",
    # 'Temperature of snow layer',
    "Total precipitation",
    # 'Volumetric soil water layer 1',
    # 'Volumetric soil water layer 2',
    # 'Volumetric soil water layer 3',
    # 'Volumetric soil water layer 4'
]


def open_dataset(
    data_variables=variables, start_time=start, end_time=end, bbox=box, time_chunks=24
):
    """
    从minio服务器读取era5-land数据

    Args:
        data_variables (list): 数据变量列表
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        bbox (list|tuple): 四至范围
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": FS.open(f"s3://{bucket_name}/geodata/era5_land/era5_land_.json"),
                "remote_protocol": "s3",
                "remote_options": RO,
            },
        },
    )

    ds = ds.filter_by_attrs(long_name=lambda v: v in data_variables)
    ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose("time", "lon", "lat")

    start_time = max(start_time, start)
    end_time = min(end_time, end)
    times = slice(start_time, end_time)
    ds = ds.sel(time=times)

    bbox = regen_box(bbox, 0.1, 0)

    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]

    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]

    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]

    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]

    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)

    ds = ds.sortby("lat", ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)

    return ds


import geopandas as gpd


def from_shp(
    data_variables=variables, start_time=start, end_time=end, shp=None, time_chunks=24
):
    """
    通过已有的矢量数据范围从minio服务器读取era5-land数据

    Args:
        data_variables (list): 数据变量列表
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        shp (str): 矢量数据路径
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    gdf = gpd.GeoDataFrame.from_file(shp)
    b = gdf.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]), 0.1, 0
    )

    return open_dataset(data_variables, start_time, end_time, bbox, time_chunks)


def from_aoi(
    data_variables=variables,
    start_time=start,
    end_time=end,
    aoi: gpd.GeoDataFrame = None,
    time_chunks=24,
):
    """
    用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取era5-land数据

    Args:
        data_variables (list): 数据变量列表
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    b = aoi.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]), 0.1, 0
    )

    return open_dataset(data_variables, start_time, end_time, bbox, time_chunks)


from netCDF4 import Dataset, date2num, num2date
import time
from datetime import datetime, timedelta


def _creatspinc(value, data_vars, lats, lons, starttime, filename, resolution):
    gridspi = Dataset(filename, "w", format="NETCDF4")

    # dimensions
    gridspi.createDimension("time", value[0].shape[0])
    gridspi.createDimension("lat", value[0].shape[2])  # len(lat)
    gridspi.createDimension("lon", value[0].shape[1])

    # Create coordinate variables for dimensions
    times = gridspi.createVariable("time", np.float64, ("time",))
    latitudes = gridspi.createVariable("lat", np.float32, ("lat",))
    longitudes = gridspi.createVariable("lon", np.float32, ("lon",))

    # Create the actual variable
    for var, attr in data_vars.items():
        gridspi.createVariable(
            var,
            np.float32,
            (
                "time",
                "lon",
                "lat",
            ),
        )

    # Global Attributes
    gridspi.description = "var"
    gridspi.history = f"Created {time.ctime(time.time())}"
    gridspi.source = "netCDF4 python module tutorial"

    # Variable Attributes
    latitudes.units = "degree_north"
    longitudes.units = "degree_east"
    times.units = "days since 1970-01-01 00:00:00"
    times.calendar = "gregorian"

    # data
    latitudes[:] = lats
    longitudes[:] = lons

    # Fill in times
    dates = []
    if resolution == "daily":
        dates.extend(starttime + n for n in range(value[0].shape[0]))
        times[:] = dates[:]

    elif resolution == "6-hourly":
        # for n in range(value[0].shape[0]):
        #     dates.append(starttime + (n+1) * np.timedelta64(6, 'h'))

        dates.extend(
            starttime + (n + 1) * timedelta(hours=6) for n in range(value[0].shape[0])
        )
        times[:] = date2num(dates, units=times.units, calendar=times.calendar)
        # print 'time values (in units %s): ' % times.units +'\n', times[:]
        dates = num2date(times[:], units=times.units, calendar=times.calendar)

    # Fill in values
    i = 0
    for var, attr in data_vars.items():
        gridspi.variables[var].long_name = attr["long_name"]
        gridspi.variables[var].units = attr["units"]
        gridspi.variables[var][:] = value[i][:]
        i = i + 1

    gridspi.close()


def to_netcdf(
    data_variables=variables,
    start_time=start,
    end_time=end,
    shp=None,
    resolution="hourly",
    save_file="era5.nc",
    time_chunks=24,
):
    """
    读取数据并保存为本地nc文件

    Args:
        data_variables (list): 数据变量列表
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        shp (str): 已有的矢量数据路径
        resolution (str): 输出的时间分辨率
        save_file (str): 输出的文件路径
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    gdf = gpd.GeoDataFrame.from_file(shp)
    b = gdf.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]), 0.1, 0
    )

    if resolution == "hourly":
        ds = open_dataset(data_variables, start_time, end_time, bbox, time_chunks)

        if ds.to_netcdf(save_file) is None:
            print(save_file, "已生成")
            ds = xr.open_dataset(save_file)
            return ds

    if resolution == "daily":
        start_time = np.datetime64(f"{str(start_time)[:10]}T01:00:00.000000000")
        end_time = np.datetime64(str(end_time)[:10]) + 1
        end_time = np.datetime64(f"{str(end_time)}T00:00:00.000000000")

        ds = open_dataset(data_variables, start_time, end_time, bbox, time_chunks)

        days = ds["time"].size // 24

        data_vars = {k: v.attrs for k, v in ds.data_vars.items()}
        daily_arr = []

        for var, attr in data_vars.items():
            a = ds[var].to_numpy()

            if attr["long_name"] in accumulated:
                xlist = [x for x in range(a.shape[0]) if x % 24 != 23]
                _a = np.delete(a, xlist, axis=0)

                daily_arr.append(_a)

            else:
                r = np.split(a, days, axis=0)
                _r = [
                    np.expand_dims(np.mean(r[i], axis=0), axis=0) for i in range(len(r))
                ]
                __r = np.concatenate(_r)

                daily_arr.append(__r)

        lats = ds["lat"].to_numpy()
        lons = ds["lon"].to_numpy()

        start_time = np.datetime64(str(start_time)[:10])

        _creatspinc(daily_arr, data_vars, lats, lons, start_time, save_file, "daily")

        new = xr.open_dataset(save_file)
        print(save_file, "已生成")
        return new

    if resolution == "6-hourly":
        start_time = np.datetime64(f"{str(start_time)[:10]}T01:00:00.000000000")
        end_time = np.datetime64(str(end_time)[:10]) + 1
        end_time = np.datetime64(f"{str(end_time)}T00:00:00.000000000")

        ds = open_dataset(data_variables, start_time, end_time, bbox, time_chunks)

        days = ds["time"].size // 6

        data_vars = {k: v.attrs for k, v in ds.data_vars.items()}
        daily_arr = []

        for var, attr in data_vars.items():
            a = ds[var].to_numpy()

            if attr["long_name"] in accumulated:
                xlist = [x for x in range(a.shape[0]) if x % 6 != 5]
                _a = np.delete(a, xlist, axis=0)

                daily_arr.append(_a)

            else:
                r = np.split(a, days, axis=0)
                _r = [
                    np.expand_dims(np.mean(r[i], axis=0), axis=0) for i in range(len(r))
                ]
                __r = np.concatenate(_r)

                daily_arr.append(__r)

        lats = ds["lat"].to_numpy()
        lons = ds["lon"].to_numpy()

        # start_time = np.datetime64(f'{str(start_time)[:10]}')
        year = int(f"{str(start_time)[:4]}")
        month = int(f"{str(start_time)[5:7]}")
        day = int(f"{str(start_time)[8:10]}")
        dt = datetime(year, month, day, 0, 0, 0)

        _creatspinc(daily_arr, data_vars, lats, lons, dt, save_file, "6-hourly")

        new = xr.open_dataset(save_file)
        print(save_file, "已生成")
        return new



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-22 13:38:07
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-08-15 17:07:37
FilePath: /hydrodatasource/tests/test_streamflow_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

import pytest
from hydrodatasource.cleaner.streamflow_cleaner import StreamflowCleaner, StreamflowBacktrack  # 确保引入你的类
import pandas as pd
import matplotlib.pyplot as plt
import glob
from tqdm import tqdm
import os

def test_anomaly_process():
    # 测试径流数据处理功能，单独处理csv文件，修改该过程可实现文件夹批处理多个文件
    cleaner = StreamflowCleaner("/ftproot/tests_stations_anomaly_detection/streamflow_cleaner/21312150.csv",window_size=7)
    # methods默认可以联合调用，也可以单独调用。大多数情况下，默认调用moving_average
    methods = ["EMA"]
    cleaner.anomaly_process(methods)
    print(cleaner.origin_df)
    print(cleaner.processed_df)
    cleaner.processed_df.to_csv("/ftproot/tests_stations_anomaly_detection/streamflow_cleaner/21312150.csv",index=False)

def test_anomaly_process_folder():
    input_folder = "/ftproot/tests_stations_anomaly_detection/streamflow_cleaner/"
    output_folder = "/ftproot/tests_stations_anomaly_detection/streamflow_cleaner/"

    # 获取输入文件夹中所有CSV文件的路径
    csv_files = glob.glob(os.path.join(input_folder, "*.csv"))

    for csv_file in tqdm(csv_files):
        try:
            # 读取并处理每个CSV文件
            cleaner = StreamflowCleaner(csv_file,window_size=7,cutoff_frequency=0.1,iterations=2,cwt_row=1)#国内window_size=14,stride=1,cutoff_frequency=0.035,time_step=1.0,iterations=3,sampling_rate=1.0,order=5,cwt_row=2,
            methods = ["EMA"]
            cleaner.anomaly_process(methods)
            
            # 确定输出文件路径
            output_file = os.path.join(output_folder, os.path.basename(csv_file))

            # 保存处理后的数据
            cleaner.processed_df.to_csv(output_file, index=False)

            print(f"Processed {csv_file} and saved to {output_file}")
        except Exception as e:
            print(f"Error processing {csv_file}: {e}")

def test_process_backtrack():
    # 测试径流数据反推处理功能
    cleaner = StreamflowBacktrack("/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/","/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/")
    cleaner.process_backtrack()

def test_delete_nan_inq():
    # 测试径流数据反推处理功能
    cleaner = StreamflowBacktrack("/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/","/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/")
    cleaner.delete_nan_inq(data_path="/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/zq_CHN_songliao_21401550.csv",
                           file="zq_CHN_songliao_21401550.csv",
                           output_folder="/ftproot/tests_stations_anomaly_detection/streamflow_backtrack/")


import os

import numpy as np
import pandas as pd
from shapely import distance


# 雨量数据异常处理, 参见https://dlut-water.yuque.com/aq3llt/iag1ec/32727520
def get_filter_data_by_time(
    data_path,
    rain_attr="DRP",
    time_attr="TM",
    id_attr="STCD",
    rain_max_hour=200,
    filter_list=None,
):
    """

    Parameters
    ----------
    data_path : _type_
        保存待处理数据文件的文件夹，姑且认为数据文件为“86_单位名称+站号.csv”形式，参考https://dlut-water.yuque.com/kgo8gd/tnld77/pum8d50qrbs1474h
    rain_attr : str, optional
        表格中标示降雨的属性（列名），默认为DRP, by default "DRP"
    time_attr : str, optional
        表格中标示时间的属性（列名），默认为TM, by default "TM"
    id_attr : str, optional
        表格中标示站号的属性（列名），默认为STCD, by default "STCD"
    rain_max_hour : int, optional
        每小时最大降雨量，默认为200，超过这个阈值的排除, by default 200
    filter_list : _type_, optional
        其他预处理过程得到的黑名单，以过滤不可用的站点, by default None

    Returns
    -------
    _type_
        _description_
    """
    if filter_list is None:
        filter_list = []
    time_df_dict = {}
    for dir_name, sub_dir, files in os.walk(data_path):
        for file in files:
            # 目前stcd采用文件名分离出的‘单位名称+站号’，和站点数据表里的站点号未必一致，需要另行考虑
            stcd = (file.split(".")[0]).split("_")[1]
            cached_csv_path = os.path.join(data_path, stcd + ".csv")
            if (stcd not in filter_list) & (~os.path.exists(cached_csv_path)):
                drop_list = []
                csv_path = os.path.join(data_path, file)
                table = pd.read_csv(csv_path, engine="c")
                # 按降雨最大阈值为200和小时雨量一致性过滤索引
                # 有些数据不严格按照小时尺度排列，出于简单可以一概按照小时重采样
                if rain_attr in table.columns:
                    table[time_attr] = pd.to_datetime(
                        table[time_attr], format="%Y-%m-%d %H:%M:%S"
                    )
                    table = table.drop(index=table.index[table[rain_attr].isna()])
                    # 整小时数据，再按小时重采样求和，结果不变
                    table = table.set_index(time_attr).resample("H").sum()
                    cached_time_array = table.index[table[id_attr] != 0].to_numpy()
                    cached_drp_array = table[rain_attr][table[id_attr] != 0].to_numpy()
                    table[time_attr] = np.nan
                    table[rain_attr][cached_time_array] = cached_drp_array
                    table = table.fillna(-1).reset_index()
                    for i in range(len(table[rain_attr])):
                        if table[rain_attr][i] > rain_max_hour:
                            drop_list.append(i)
                        if i >= 5:
                            hour_slice = table[rain_attr][i - 5 : i + 1].to_numpy()
                            if hour_slice.all() == np.mean(hour_slice):
                                drop_list.extend(list(range(i - 5, i + 1)))
                    drop_array = np.unique(np.array(drop_list, dtype=int))
                    table = table.drop(index=drop_array)
                    drop_array_minus = table.index[table[time_attr] == -1]
                    table = table.drop(index=drop_array_minus)
                time_df_dict[stcd] = table
                # 在这里会生成csv数据表，存在副作用
                table.to_csv(cached_csv_path)
            elif (int(stcd) not in filter_list) & (os.path.exists(cached_csv_path)):
                table = pd.read_csv(cached_csv_path, engine="c")
                time_df_dict[stcd] = table
    return time_df_dict


def get_filter_data_by_space(
    time_df_dict,
    filter_list,
    station_gdf,
    csv_path,
    rain_attr="DRP",
    time_attr="TM",
    id_attr="STCD",
):
    """
    :param time_df_dict: 根据get_filter_data_by_time()得到的中间数据dict
    :param filter_list: 其他预处理过程得到的黑名单，以过滤不可用的站点
    :param station_gdf: 存储站点位置的GeoDataFrame
    :param csv_path: 按站点保存清洗后数据的文件夹
    :param rain_attr: 表格中标示降雨的属性（列名），默认为DRP
    :param time_attr: 表格中标示时间的属性（列名），默认为TM
    :param id_attr: station_gdf表格中标示站号的属性（列名），默认为STCD
    :return: 站点号与根据空间均值排除后的表格构成的dict
    """
    neighbor_stas_dict = find_neighbor_dict(station_gdf, filter_list)[0]
    space_df_dict = {}
    for key in time_df_dict:
        time_drop_list = []
        neighbor_stas = neighbor_stas_dict[key]
        table = time_df_dict[key]
        table = table.set_index(time_attr)
        for time in table.index:
            rain_time_dict = {}
            for neighbor in neighbor_stas:
                neighbor_df = time_df_dict[str(neighbor)]
                neighbor_df = neighbor_df.set_index(time_attr)
                if time in neighbor_df.index:
                    rain_time_dict[str(neighbor)] = neighbor_df[rain_attr][time]
            if not rain_time_dict:
                continue
            elif 0 < len(rain_time_dict) < 12:
                weight_rain = 0
                weight_dis = 0
                for sta in rain_time_dict:
                    point = station_gdf.geometry[
                        station_gdf[id_attr] == str(sta)
                    ].values[0]
                    point_self = station_gdf.geometry[
                        station_gdf[id_attr] == str(key)
                    ].values[0]
                    dis = distance(point, point_self)
                    weight_rain += table[rain_attr][time] / (dis**2)
                    weight_dis += 1 / (dis**2)
                interp_rain = weight_rain / weight_dis
                if abs(interp_rain - table[rain_attr][time]) > 4:
                    time_drop_list.append(time)
            elif len(rain_time_dict) >= 12:
                rain_time_series = pd.Series(rain_time_dict.values())
                quantile_25 = rain_time_series.quantile(q=0.25)
                quantile_75 = rain_time_series.quantile(q=0.75)
                average = rain_time_series.mean()
                if rain_attr in table.columns:
                    MA_Tct = (table[rain_attr][time] - average) / (
                        quantile_75 - quantile_25
                    )
                    if MA_Tct > 4:
                        time_drop_list.append(time)
        table = table.drop(index=time_drop_list).drop(columns=["Unnamed: 0"])
        space_df_dict[key] = table
        # 会生成csv文件，有副作用
        table.to_csv(os.path.join(csv_path, key + ".csv"))
    return space_df_dict


def find_neighbor_dict(station_gdf, filter_list, id_attr="STCD"):
    """
    :param station_gdf: 存储有站点位置的GeoDataFrame
    :param filter_list: 其他预处理过程得到的黑名单，以过滤不可用的站点
    :param id_attr: station_gdf表格中标示站号的属性（列名），默认为STCD
    :return: 与各站相邻的站点号（取0-0.2度）
    """
    station_gdf = station_gdf.set_index(id_attr).drop(index=filter_list).reset_index()
    station_gdf[id_attr] = station_gdf[id_attr].astype("str")
    neighbor_dict = {}
    for i in range(len(station_gdf.geometry)):
        stcd = station_gdf[id_attr][i]
        station_gdf["distance"] = station_gdf.apply(
            lambda x: distance(station_gdf.geometry[i], x.geometry), axis=1
        )
        nearest_stas = station_gdf[
            (station_gdf["distance"] > 0) & (station_gdf["distance"] <= 0.2)
        ]
        nearest_stas_list = nearest_stas[id_attr].to_list()
        neighbor_dict[stcd] = nearest_stas_list
    station_gdf = station_gdf.drop(columns=["distance"])
    return neighbor_dict



"""
Author: Wenyu Ouyang
Date: 2023-11-01 08:58:50
LastEditTime: 2024-06-25 09:42:12
LastEditors: Wenyu Ouyang
Description: Test funcs for reader.py
FilePath: \hydrodatasource\tests\test_reader.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
from minio import Minio
import hydrodataset as hds

from hydrodatasource.configs.config import LOCAL_DATA_PATH
from hydrodatasource.reader.reader import (
    AOI,
    DataHandler,
    LocalFileReader,
    MinioFileReader,
)


def test_gpm_read_minio():
    data_handler = DataHandler(
        aoi_type="basin",  # only basin for now, the streamflow data needs to be reorganized in the minio server,  then to be read.
        aoi_param="86_21401550",
        dataname="gpm",
        minio_read=True,
    )
    data = data_handler.handle()
    assert data is not None
    return data


def test_gfs_read_minio():
    data_handler = DataHandler(
        aoi_type="basin", aoi_param="86_21401550", dataname="gfs", minio_read=True
    )
    data = data_handler.handle()
    assert data is not None
    return data


def test_gpm_gfs_read_minio():
    data_handler = DataHandler(
        aoi_type="basin", aoi_param="86_21401550", dataname="gpm_gfs", minio_read=True
    )
    data = data_handler.handle()
    assert data is not None
    return data


def test_gpm_read_with_process():
    # if there is no data in minio, we want to merge a new one and save in local_data_path
    data_handler = DataHandler(
        aoi_type="basin",
        aoi_param="86_21401550",
        region="wis",
        time_periods=[["2017-01-01T00:00:00", "2017-01-31T00:00:00"]],
        dataname="gpm",
        minio_read=False,
        local_save=True,
    )
    data = data_handler.handle()
    assert data is not None
    return data


# The new method no longer needs ways below, but I keep them in case we need it.

# def test_reader_interface(minio_paras):
#     # 初始化Minio客户端
#     minio_server = minio_paras["endpoint_url"]
#     minio_client = Minio(
#         minio_server.replace("http://", ""),
#         access_key=minio_paras["access_key"],
#         secret_key=minio_paras["secret_key"],
#         secure=False,
#     )

#     gpm_handler = GPMDataHandler()
#     gfs_handler = GFSDataHandler()
#     aoi = AOI("grid", {"lat": 0, "lon": 0, "size": 1})

#     local_gpm_reader = LocalFileReader(gpm_handler)
#     local_gfs_reader = LocalFileReader(gfs_handler)
#     local_gpm_reader.read("path/to/file", aoi)

#     # Assume you have initialized the minio_client somewhere
#     minio_gpm_reader = MinioFileReader(minio_client, gpm_handler)
#     minio_gfs_reader = MinioFileReader(minio_client, gfs_handler)
#     minio_gpm_reader.read("path/to/file", aoi)


# def test_reader_station():
#     station_handler = StationDataHandler()
#     aoi = AOI(
#         "station",
#         {"station_id": "2181200", "start_time": "1980-01-01", "end_time": "2001-01-01"},
#     )

#     local_station_reader = LocalFileReader(station_handler)
#     data = local_station_reader.read(
#         os.path.join(LOCAL_DATA_PATH, "station.nc"),
#         aoi,
#     )
#     print(data)



"""
Author: Wenyu Ouyang
Date: 2024-07-06 19:20:59
LastEditTime: 2024-09-14 14:58:28
LastEditors: Wenyu Ouyang
Description: Test funcs for data source
FilePath: \hydrodatasource\tests\test_data_source.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from hydrodatasource.configs.config import SETTING
from hydrodatasource.reader.data_source import CACHE_DIR, SelfMadeHydroDataset


@pytest.fixture
def one_hour_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(data_path=selfmadehydrodataset_path, time_unit=["1h"])


@pytest.fixture
def three_hour_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(data_path=selfmadehydrodataset_path, time_unit=["3h"])


@pytest.fixture
def one_day_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(data_path=selfmadehydrodataset_path)


@pytest.fixture
def eight_day_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(data_path=selfmadehydrodataset_path, time_unit=["8D"])


def test_selfmadehydrodataset_get_name(one_day_dataset):
    assert one_day_dataset.get_name() == "SelfMadeHydroDataset"


def test_selfmadehydrodataset_streamflow_unit(one_day_dataset):
    assert one_day_dataset.streamflow_unit == {"1D": "mm/d"}


def test_selfmadehydrodataset_read_site_info(one_day_dataset):
    site_info = one_day_dataset.read_site_info()
    assert isinstance(site_info, pd.DataFrame)


def test_selfmadehydrodataset_read_object_ids(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    assert isinstance(object_ids, np.ndarray)


def test_selfmadehydrodataset_read_tsdata(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    target_cols = one_day_dataset.read_timeseries(
        object_ids=object_ids[:5],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["streamflow"],
        time_unit=["1D"],
    )
    assert isinstance(target_cols, dict)


def test_selfmadehydrodataset_read_attrdata(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    constant_cols = one_day_dataset.read_attributes(
        object_ids=object_ids[:5], constant_cols=["area"]
    )
    assert isinstance(constant_cols, np.ndarray)


def test_selfmadehydrodataset_get_attributes_cols(one_day_dataset):
    constant_cols = one_day_dataset.get_attributes_cols()
    assert isinstance(constant_cols, np.ndarray)


def test_selfmadehydrodataset_get_timeseries_cols(one_day_dataset):
    relevant_cols = one_day_dataset.get_timeseries_cols()
    assert isinstance(relevant_cols, dict)


def test_selfmadehydrodataset_cache_attributes_xrdataset(one_day_dataset):
    one_day_dataset.cache_attributes_xrdataset()
    assert os.path.exists(os.path.join(CACHE_DIR, "attributes.nc"))


def test_selfmadehydrodataset_cache_timeseries_xrdataset(
    one_day_dataset, three_hour_dataset, one_hour_dataset, eight_day_dataset
):
    # 8D
    eight_day_dataset.cache_timeseries_xrdataset(
        time_units=["8D"],
        t_range=["1980-01-01", "2023-12-31"],
        start0101_freq=True,
        batchsize=200,
    )
    # 1h
    one_hour_dataset.cache_timeseries_xrdataset(
        time_units=["1h"],
        t_range=["1980-01-01", "2023-12-31"],
    )
    # 3h
    three_hour_dataset.cache_timeseries_xrdataset(
        time_units=["3h"],
        t_range=["1980-01-01 01", "2023-12-31 22"],
    )
    # 1D
    one_day_dataset.cache_timeseries_xrdataset()


def test_selfmadehydrodataset_cache_xrdataset(one_day_dataset):
    one_day_dataset.cache_xrdataset()


def test_selfmadehydrodataset_read_ts_xrdataset(
    one_day_dataset, three_hour_dataset, one_hour_dataset, eight_day_dataset
):
    # 8D
    xrdataset_dict = eight_day_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31"],
        var_lst=["ET_modis16a2006", "ET_modis16a2gf061"],
        time_units=["8D"],
    )
    target_cols = one_day_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["streamflow"],
        time_unit=["1D"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["1D"]["streamflow"].values, target_cols["1D"][:, :, 0]
    )
    # 1h
    xrdataset_dict = one_hour_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31 23"],
        var_lst=["streamflow"],
        time_units=["1h"],
    )
    target_cols = one_hour_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31 23"],
        relevant_cols=["streamflow"],
        time_units=["1h"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["1h"]["streamflow"].values, target_cols["1h"][:, :, 0]
    )

    # 3h
    xrdataset_dict = three_hour_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01 01", "2020-12-31 22"],
        var_lst=["streamflow"],
        time_units=["3h"],
    )
    target_cols = three_hour_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01 01", "2020-12-31 22"],
        relevant_cols=["streamflow"],
        time_units=["3h"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["3h"]["streamflow"].values, target_cols["3h"][:, :, 0]
    )

    # 1D
    xrdataset_dict = one_day_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31"],
        var_lst=["streamflow"],
        time_units=["1D"],
    )
    target_cols = one_day_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["streamflow"],
        time_unit=["1D"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["1D"]["streamflow"].values, target_cols["1D"][:, :, 0]
    )


def test_selfmadehydrodataset_read_attr_xrdataset(one_day_dataset):
    xrdataset = one_day_dataset.read_attr_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        var_lst=["area"],
    )
    assert isinstance(xrdataset, xr.Dataset)


def test_selfmadehydrodataset_read_area(one_day_dataset):
    area = one_day_dataset.read_area(gage_id_lst=["camels_01013500", "camels_01022500"])
    assert isinstance(area, xr.Dataset)


def test_selfmadehydrodataset_read_mean_prcp(one_day_dataset):
    mean_prcp = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"]
    )
    assert isinstance(mean_prcp, xr.Dataset)



#!/usr/bin/env python
"""
Author: Wenyu Ouyang
Date: 2023-10-25 15:16:21
LastEditTime: 2024-03-28 08:40:25
LastEditors: Wenyu Ouyang
Description: Tests for reading basins' data
FilePath: \hydrodatasource\tests\test_reader_basins.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import os.path
import pathlib

from hydrodatasource.reader import stations
from hydrodatasource.configs import config
from hydrodatasource.reader import minio_api


def test_content():
    # preprocess.huanren_preprocess()
    stations.biliu_stbprp_decode()


def test_upload_csv():
    test_path = pathlib.Path(os.path.abspath(os.path.curdir)).parent
    client = config.MC
    bucket_name = config.MINIO_PARAM["bucket_path"]
    minio_api.boto3_upload_csv(
        client,
        bucket_name,
        "nyc_taxi",
        os.path.join(test_path, "test_data/nyc_taxi.csv"),
    )
    minio_api.minio_upload_csv(
        client,
        bucket_name,
        "driver_data_site24",
        os.path.join(test_path, "test_data/driver_data_site24.csv"),
    )


def test_download_csv_minio():
    client = config.S3
    bucket_name = config.MINIO_PARAM["bucket_path"]
    # minio_api.minio_download_csv(client, bucket_name, 'nyc_taxi', file_path='test_dload')
    minio_api.boto3_download_csv(
        client, bucket_name, "driver_data_site24", "driver_data_site24.csv"
    )



import os

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
import xarray as xr
import tzfpy
import dataretrieval.nwis as nwis

from hydrodatasource.reader.spliter_grid import read_streamflow_from_minio
from hydrodatasource.configs.config import FS


def test_gen_camels_hourly_shp():
    camels_shp = '/ftproot/camels/camels_us/basin_set_full_res/HCDN_nhru_final_671.shp'
    camels_hourly_csvs = '/ftproot/camels_hourly/data/usgs_streamflow_csv/'
    csv_paths = os.listdir(camels_hourly_csvs)
    minio_csv_paths = FS.glob('s3://basins-origin/basin_shapefiles/**')[1:]
    minio_stcds = [csv.split('_')[-1].split('.')[0] for csv in minio_csv_paths]
    hourly_basin_ids = [path.split('-')[0] for path in csv_paths]
    camels_gpd = gpd.read_file(camels_shp)
    camels_gpd['hru_id'] = camels_gpd['hru_id'].astype('str')
    camels_gpd['hru_id'] = camels_gpd['hru_id'].apply(lambda x: x.zfill(8))
    camels_hourly_gpd_blank = camels_gpd[camels_gpd['hru_id'].isin(hourly_basin_ids)]
    camels_hourly_gpd = camels_hourly_gpd_blank[~camels_gpd['hru_id'].isin(minio_stcds)]
    camels_hourly_gpd.to_file('/home/jiaxuwu/camels_hourly_basins.shp')


def test_read_usa_streamflow():
    # basin_list = hdscc.FS.glob('s3://basins-origin/basin_shapefiles/**')
    basin_list = ["basin_USA_camels_01181000", "basin_USA_camels_01411300",
                  "basin_USA_camels_01414500", "basin_USA_camels_02016000",
                  "basin_USA_camels_02018000", "basin_USA_camels_02481510",
                  "basin_USA_camels_03070500", "basin_USA_camels_08324000",
                  "basin_USA_camels_11266500", "basin_USA_camels_11523200",
                  "basin_USA_camels_12020000", "basin_USA_camels_12167000",
                  "basin_USA_camels_14185000", "basin_USA_camels_14306500",
                  "basin_CHN_songliao_21401550", "basin_USA_camels_14400000"]
    for basin_id in basin_list:
        s3_basin_path = f's3://basins-origin/basin_shapefiles/{basin_id}.zip'
        basin_gpd = gpd.read_file(FS.open(s3_basin_path))
        basin_tz = tzfpy.get_tz(basin_gpd.geometry[0].centroid.x, basin_gpd.geometry[0].centroid.y)
        q_array = read_streamflow_from_minio(times=[['2014-12-31 17:00:00', '2019-12-31 23:00:00'],
                                          ['2020-01-01 00:00:00', '2023-12-31 23:00:00']],
                                   sta_id=basin_id.lstrip('basin_'))
        if basin_tz == 'America/Los_Angeles':
            q_array['TM'] = q_array['TM'] + np.timedelta64(7, 'h')
        elif basin_tz == 'America/Denver':
            q_array['TM'] = q_array['TM'] + np.timedelta64(6, 'h')
        elif basin_tz == 'America/Chicago':
            q_array['TM'] = q_array['TM'] + np.timedelta64(5, 'h')
        elif basin_tz == 'America/New_York':
            q_array['TM'] = q_array['TM'] + np.timedelta64(4, 'h')
        q_array_2019 = q_array[q_array['TM'] < '2020-01-01 00:00:00']
        q_array_2020 = q_array[q_array['TM'] >= '2020-01-01 00:00:00']
        # convert degree^2 to km^2
        basin_area = gpd.read_file(FS.open(s3_basin_path)).geometry[0].area * 12345.6789
        # convert ft^3 to m^3
        q_array_2020['Q'] = q_array_2020['Q'] / 35.31
        # convert m^3 to mm/h
        q_array_2020['Q'] = q_array_2020['Q'] / basin_area * 3.6
        q_array_mm_h = pd.concat([q_array_2019, q_array_2020])
        q_ds = xr.Dataset.from_dataframe(q_array_mm_h.set_index('TM'))
        q_ds.to_netcdf(basin_id+'_streamflow.nc')


def test_download_from_usgs():
    '''
    camels_hourly_csvs = '/ftproot/camels_hourly/data/usgs_streamflow_csv/'
    csv_paths = os.listdir(camels_hourly_csvs)
    minio_csv_paths = FS.glob('s3://basins-origin/basin_shapefiles/**')[1:]
    hourly_basin_ids = [path.split('-')[0] for path in csv_paths]
    minio_stcds = [csv.split('_')[-1].split('.')[0] for csv in minio_csv_paths]
    empty_stcds = [stcd for stcd in hourly_basin_ids if stcd not in minio_stcds]
    '''
    # usgs_basins = gpd.read_file("/ftproot/usgs_camels_hourly_flowdb_1373/concat_flow_camels_1373.zip")
    # first_empty_stcds = usgs_basins['basin_id'].to_numpy()
    usgs_gages = gpd.read_file("/ftproot/usgs_camels_hourly_flowdb_1373/concat_usa_usgs_all.shp")
    empty_stcds = usgs_gages['GAGE_ID'].to_numpy()
    # empty_stcds = np.union1d(first_empty_stcds, second_empty_stcds)
    for site in empty_stcds:
        site_path = f'/ftproot/usgs_camels_hourly_flowdb_1373/zq_USA_usgs_{site}.csv'
        if not os.path.exists(site_path):
            try:
                site_df = nwis.get_record(sites=site, service='iv', start='2015-01-01')
            except requests.exceptions.ConnectionError:
                continue
            except requests.exceptions.JSONDecodeError:
                continue
            site_df.to_csv(site_path)




"""
Author: Wenyu Ouyang
Date: 2024-03-23 15:10:23
LastEditTime: 2024-05-20 18:09:22
LastEditors: Wenyu Ouyang
Description: Test for utility functions
FilePath: \hydrodatasource\tests\test_utils.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import numpy as np
import xarray as xr
import pint
import pytest

from hydrodatasource.utils.utils import (
    streamflow_unit_conv,
    minio_file_list,
    is_minio_folder,
)

ureg = pint.UnitRegistry()
ureg.force_ndarray_like = True  # or ureg.force_ndarray = True


# Test case for xarray input
@pytest.mark.parametrize(
    "streamflow, area, target_unit, inverse, expected",
    # [
    #     (
    #         xr.Dataset(
    #             {
    #                 "streamflow": xr.DataArray(
    #                     np.array([[100, 200], [300, 400]]), dims=["time", "basin"]
    #                 )
    #             }
    #         ),
    #         xr.Dataset({"area": xr.DataArray(np.array([1, 2]), dims=["basin"])}),
    #         "mm/d",
    #         False,
    #         xr.Dataset(
    #             {
    #                 "streamflow": xr.DataArray(
    #                     np.array(
    #                         [
    #                             [8640.0, 8640.0],
    #                             [25920.0, 17280.0],
    #                         ]
    #                     ),
    #                     dims=["time", "basin"],
    #                 )
    #             }
    #         ),
    #     ),
    # ],
    # Add more test cases for xarray input
    [
        (
            xr.Dataset(
                {
                    "streamflow": xr.DataArray(
                        np.array([[100, 200], [300, 400]]), dims=["time", "basin"]
                    )
                }
            ),
            xr.Dataset({"area": xr.DataArray(np.array([1, 2]), dims=["basin"])}),
            "mm/3h",
            False,
            xr.Dataset(
                {
                    "streamflow": xr.DataArray(
                        np.array(
                            [
                                [1080.0, 1080.0],
                                [3240.0, 2160.0],
                            ]
                        ),
                        dims=["time", "basin"],
                    )
                }
            ),
        ),
    ],
)
def test_streamflow_unit_conv_xarray(streamflow, area, target_unit, inverse, expected):
    # Attaching units using pint
    streamflow["streamflow"] = streamflow["streamflow"] * ureg.m**3 / ureg.s
    area["area"] = area["area"] * ureg.km**2

    result = streamflow_unit_conv(streamflow, area, target_unit, inverse)
    xr.testing.assert_allclose(result, expected)


@pytest.mark.parametrize(
    "streamflow, area, target_unit, inverse, expected",
    # [
    #     (
    #         xr.Dataset(
    #             {
    #                 "streamflow": xr.DataArray(
    #                     np.array(
    #                         [
    #                             [8640.0, 8640.0],
    #                             [25920.0, 17280.0],
    #                         ]
    #                     ),
    #                     dims=["time", "basin"],
    #                     attrs={"units": "mm/d"},
    #                 )
    #             }
    #         ),
    #         xr.Dataset(
    #             {
    #                 "area": xr.DataArray(
    #                     np.array([1, 2]), dims=["basin"], attrs={"units": "km^2"}
    #                 )
    #             }
    #         ),
    #         "m^3/s",
    #         True,
    #         xr.Dataset(
    #             {
    #                 "streamflow": xr.DataArray(
    #                     np.array([[100, 200], [300, 400]]),
    #                     dims=["time", "basin"],
    #                 )
    #             }
    #         ),
    #     ),
    # ],
    [
        (
            xr.Dataset(
                {
                    "streamflow": xr.DataArray(
                        np.array(
                            [
                                [1080.0, 1080.0],
                                [3240.0, 2160.0],
                            ]
                        ),
                        dims=["time", "basin"],
                        attrs={"units": "mm/3h"},
                    )
                }
            ),
            xr.Dataset(
                {
                    "area": xr.DataArray(
                        np.array([1, 2]), dims=["basin"], attrs={"units": "km^2"}
                    )
                }
            ),
            "m^3/s",
            True,
            xr.Dataset(
                {
                    "streamflow": xr.DataArray(
                        np.array([[100, 200], [300, 400]]),
                        dims=["time", "basin"],
                    )
                }
            ),
        ),
    ],
)
def test_streamflow_unit_conv_xarray_inverse(
    streamflow, area, target_unit, inverse, expected
):
    result = streamflow_unit_conv(streamflow, area, target_unit, inverse)
    xr.testing.assert_allclose(result, expected)


# Test case for numpy and pandas input
@pytest.mark.parametrize(
    "streamflow, area, target_unit, inverse, expected",
    # [
    #     (
    #         np.array([100, 200]) * ureg.m**3 / ureg.s,
    #         np.array([1]) * ureg.km**2,
    #         "mm/d",
    #         False,
    #         np.array([8640.0, 17280.0]),
    #     ),
    # ],
    # [
    #     (
    #         np.array([100, 200]) * ureg.m**3 / ureg.s,
    #         np.array([1]) * ureg.km**2,
    #         "mm/3h",
    #         False,
    #         np.array([1080.0, 2160.0]),
    #     ),
    # ],
    # [
    #     (
    #         np.array([8640.0, 17280.0]) * ureg.mm / ureg.d,
    #         np.array([1]) * ureg.km**2,
    #         "m^3/s",
    #         True,
    #         np.array([100, 200]),
    #     ),
    # ],
    [
        (
            np.array([1080.0, 2160.0]) * ureg.mm / ureg.h / 3,
            np.array([1]) * ureg.km**2,
            "m^3/s",
            True,
            np.array([100, 200]),
        ),
    ],
)
def test_streamflow_unit_conv_np_pd(streamflow, area, target_unit, inverse, expected):
    result = streamflow_unit_conv(streamflow, area, target_unit, inverse)
    np.testing.assert_array_almost_equal(result, expected)


# Test case for invalid input type
@pytest.mark.parametrize(
    "streamflow, area, target_unit, inverse",
    [
        (None, np.array([2, 2, 2]), "mm/d", False),
        (np.array([10, 20, 30]), None, "mm/d", False),
        (np.array([10, 20, 30]), np.array([2, 2, 2]), "invalid_unit", False),
    ],
)
def test_streamflow_unit_conv_invalid_input(streamflow, area, target_unit, inverse):
    with pytest.raises(TypeError):
        streamflow_unit_conv(streamflow, area, target_unit, inverse)


def test_minio_file_list():
    minio_folder_url = "s3://basins-interim/timeseries/1D"
    file_list = minio_file_list(minio_folder_url)
    print(file_list)


def test_is_minio_folder():
    minio_folder_url = "s3://basins-interim/timeseries/1D"
    print(is_minio_folder(minio_folder_url))
    minio_folder_url = "s3://basins-interim/timeseries/1D_units_info.json"
    print(is_minio_folder(minio_folder_url))



from hydrodatasource.reader.postgres import (
    read_forcing_dataframe,
    read_plcd
)
import hydrodatasource.configs.config as hdscc
import pandas as pd


def test_read_gpm():
    data = read_forcing_dataframe("gpm_tp", "21401550", ["2024-05-19 06:00:00", None])
    print(data[['predictdate', 'tp']])


def test_read_gfs_tp():
    data = read_forcing_dataframe(
        "gfs_tp", "21401550", ["2024-05-19 06:00:00", "2024-05-19 07:00:00"]
    )
    print(data[['forecastdatetime', 'tp']])


def test_read_smap():
    data = read_forcing_dataframe(
        "smap_sm_surface", "21401550", ["2024-05-02 00:00:00", "2024-05-03 00:00:00"]
    )
    print(data[['predictdate', 'sm_surface']])


def test_read_gfs_sm():
    data = read_forcing_dataframe(
        "gfs_soilw", "21401550", ["2024-06-01 00:00:00", "2024-06-02 00:00:00"]
    )
    print(data[['forecastdatetime', 'sm_surface']])


def test_read_sl_pg():
    import geopandas as gpd
    # 获取water数据库下所有表
    all_tables = pd.read_sql(
        "SELECT table_name FROM information_schema.tables WHERE table_type = 'BASE TABLE' "
        "AND table_catalog='water'",
        hdscc.PS,
    )
    # 获取基础表信息
    stbprp_df = pd.read_sql("select * FROM ST_STBPRP_B", hdscc.PS)
    print(all_tables)
    print(stbprp_df)
    '''
    stnames = ['猴山', '石门', '土门子', '三湾', '铁甲', '太平湾水电站', '水丰', '双岭', '南城子', '榛子岭', '丰满', '杨木',
               '云峰水电站', '红石', '白山', '双沟', '小山', '松山', '向海', '老龙口']
    stnames_df = stbprp_df[['STCD', 'STNAME', 'LGTD', 'LTTD']][stbprp_df['STNM'].str.contains('|'.join(stnames))]
    # 若点处于多个流域的出口，取级别最低的那一个
    geo_column = gpd.points_from_xy(stnames_df['LGTD'], stnames_df['LTTD'])
    sta_gdf = gpd.GeoDataFrame(stnames_df).set_geometry(geo_column)
    '''

def test_read_plcd():
    basin_code = '10513040'
    data = read_plcd(basin_code)
    print(data)
    print(type(data))


"""
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-05-15 10:26:29
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-05-15 10:27:39
FilePath: /liutianxv1/hydrodatasource/tests/test_dmca_esr.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
"""


import os
import pandas as pd

import hydrodataset as hds

from hydrodatasource.cleaner.dmca_esr import *

def test_biliu_rain_flow_division():
    # rain和flow之间的索引要尽量“对齐”
    # 2014.1.1 00:00:00-2022.9.1 00:00:00
    # 读取并处理 rain.csv 文件
    filtered_rain_aver_df = pd.read_csv(
        os.path.join("/home/liutianxv1/0515场次划分/2010-2023流域平均数据.csv"),
        engine="c",
    ).set_index("TM")

    # 将时间列转换为 datetime 对象，并去重
    filtered_rain_aver_df.index = pd.to_datetime(filtered_rain_aver_df.index, errors='coerce')
    filtered_rain_aver_df = filtered_rain_aver_df[~filtered_rain_aver_df.index.duplicated(keep='first')]

    filtered_rain_aver_array = filtered_rain_aver_df["rain"].to_numpy()

    # 读取并处理 inq.csv 文件
    flow_mm_h_df = pd.read_csv(
        os.path.join("/home/liutianxv1/0515场次划分/zq_CHN_songliao_21401550.csv"),
        engine="c",
    ).set_index("TM")

    # 将时间列转换为 datetime 对象，并去重
    flow_mm_h_df.index = pd.to_datetime(flow_mm_h_df.index, errors='coerce')
    flow_mm_h_df = flow_mm_h_df[~flow_mm_h_df.index.duplicated(keep='first')]

    flow = flow_mm_h_df["INQ"]
    biliu_area = 2097
    flow_mm_h = flow.apply(lambda x: x * 3.6 / biliu_area)

    # 确保两个 DataFrame 的索引对齐
    common_index = filtered_rain_aver_df.index.intersection(flow_mm_h_df.index)
    filtered_rain_aver_df = filtered_rain_aver_df.loc[common_index]
    flow_mm_h_df = flow_mm_h_df.loc[common_index]

    time = filtered_rain_aver_df.index

    # 如果需要将时间格式重新转换为字符串，可以使用strftime
    time = pd.to_datetime(time).strftime("%Y-%m-%d %H:%M:%S")


    rain_min = 0.02
    max_window = 100
    Tr, fluct_rain_Tr, fluct_flow_Tr, fluct_bivariate_Tr = (
        step1_step2_tr_and_fluctuations_timeseries(
            filtered_rain_aver_array, flow_mm_h, rain_min, max_window
        )
    )

    beginning_core, end_core = step3_core_identification(fluct_bivariate_Tr)
    end_rain = step4_end_rain_events(
        beginning_core, end_core, filtered_rain_aver_array, fluct_rain_Tr, rain_min
    )
    beginning_rain = step5_beginning_rain_events(
        beginning_core, end_rain, filtered_rain_aver_array, fluct_rain_Tr, rain_min
    )
    beginning_rain_checked, end_rain_checked, beginning_core, end_core = (
        step6_checks_on_rain_events(
            beginning_rain,
            end_rain,
            filtered_rain_aver_array,
            rain_min,
            beginning_core,
            end_core,
        )
    )
    end_flow = step7_end_flow_events(
        end_rain_checked,
        beginning_core,
        end_core,
        filtered_rain_aver_array,
        fluct_rain_Tr,
        fluct_flow_Tr,
        Tr,
    )
    beginning_flow = step8_beginning_flow_events(
        beginning_rain_checked,
        end_rain_checked,
        filtered_rain_aver_array,
        beginning_core,
        fluct_rain_Tr,
        fluct_flow_Tr,
    )
    (
        beginning_rain_ungrouped,
        end_rain_ungrouped,
        beginning_flow_ungrouped,
        end_flow_ungrouped,
    ) = step9_checks_on_flow_events(
        beginning_rain_checked,
        end_rain_checked,
        beginning_flow,
        end_flow,
        fluct_flow_Tr,
    )
    BEGINNING_RAIN, END_RAIN, BEGINNING_FLOW, END_FLOW = (
        step10_checks_on_overlapping_events(
            beginning_rain_ungrouped,
            end_rain_ungrouped,
            beginning_flow_ungrouped,
            end_flow_ungrouped,
            time,
        )
    )
    print(len(BEGINNING_RAIN), len(END_RAIN), len(BEGINNING_FLOW), len(END_FLOW))
    # print('_________________________')
    # print('_________________________')
    # print(BEGINNING_FLOW, END_FLOW)
    multiple = 1
    flag = 0
    flow_max = 100
    biliu_area = 2097
    flow_max_mm_h = flow_max * 3.6 / biliu_area
    duration_max = 2400
    (
        DURATION_RAIN,
        VOLUME_RAIN,
        DURATION_RUNOFF,
        VULUME_RUNOFF,
        RUNOFF_RATIO,
        BEGINNING_RAIN,
        END_RAIN,
        BEGINNING_FLOW,
        END_FLOW,
    ) =  event_analysis(
        BEGINNING_RAIN,
        END_RAIN,
        BEGINNING_FLOW,
        END_FLOW,
        filtered_rain_aver_array,
        flow_mm_h,
        time,
        flag,
        multiple,
        flow_max_mm_h,
        duration_max,
    )
    biliu_division = pd.DataFrame(
        {
            "BEGINNING_RAIN": BEGINNING_RAIN,
            "END_RAIN": END_RAIN,
            "DURATION_RAIN": DURATION_RAIN,
            "BEGINNING_FLOW": BEGINNING_FLOW,
            "END_FLOW": END_FLOW,
            "DURATION_RUNOFF": DURATION_RUNOFF,
            "VOLUME_RAIN": VOLUME_RAIN,
            "VULUME_RUNOFF": VULUME_RUNOFF,
            "RUNOFF_RATIO": RUNOFF_RATIO,
        }
    )

    biliu_division.to_csv(
        os.path.join(
            "/home/liutianxv1/0515场次划分/biliu_flow_division.csv",
        ),
        index=False,
    )



"""
Author: Jingyi Wang, Yang Wang, and Wenyu Ouyang
Date: 2023-10-28 09:23:22
LastEditTime: 2024-05-20 20:52:04
LastEditors: Wenyu Ouyang
Description: DMCA-ESR method (https://doi.org/10.1029/2021WR031283)
FilePath: \hydrodatasource\hydrodatasource\cleaner\dmca_esr.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import numpy as np
import pandas as pd


def movmean(X, n):
    ones = np.ones(X.shape)
    kernel = np.ones(n)
    return np.convolve(X, kernel, mode="same") / np.convolve(ones, kernel, mode="same")


def step1_step2_tr_and_fluctuations_timeseries(rain, flow, rain_min, max_window):
    """
    The first two steps are used for calculation of the catchment response time,
    the rainfall and streamflow fluctuations, and rainfall fluctuation.

    Parameters
    ----------
    rain : np.ndarray
        unit is mm/h
    flow : np.ndarray
        unit is m³/h
    rain_min : _type_
        最小降雨量阈值
    max_window : _type_
        场次划分最大窗口, 决定场次长度

    Returns
    -------
    _type_
        _description_
    """
    rain = rain.T
    flow = flow.T
    rain_int = np.nancumsum(rain)
    flow_int = np.nancumsum(flow)
    T = rain.size
    rain_mean = np.empty(((max_window - 1) // 2, T))
    flow_mean = np.empty(((max_window - 1) // 2, T))
    fluct_rain = np.empty(((max_window - 1) // 2, T))
    fluct_flow = np.empty(((max_window - 1) // 2, T))
    F_rain = np.empty((max_window - 1) // 2)
    F_flow = np.empty((max_window - 1) // 2)
    F_rain_flow = np.empty((max_window - 1) // 2)
    rho = np.empty((max_window - 1) // 2)
    for window in np.arange(3, max_window + 1, 2):
        int_index = int((window - 1) / 2 - 1)
        start_slice = int(window - 0.5 * (window - 1))
        dst_slice = int(T - 0.5 * (window - 1))
        # 新建一个循环体长度*数据长度的大数组
        rain_mean[int_index] = movmean(rain_int, window)
        flow_mean[int_index] = movmean(flow_int, window)
        fluct_rain[int_index] = rain_int - rain_mean[int_index, :]
        F_rain[int_index] = (1 / (T - window + 1)) * np.nansum(
            (fluct_rain[int_index, start_slice:dst_slice]) ** 2
        )
        fluct_flow[int_index, np.newaxis] = flow_int - flow_mean[int_index, :]
        F_flow[int_index] = (1 / (T - window + 1)) * np.nansum(
            (fluct_flow[int_index, start_slice:dst_slice]) ** 2
        )
        F_rain_flow[int_index] = (1 / (T - window + 1)) * np.nansum(
            (fluct_rain[int_index, start_slice:dst_slice])
            * (fluct_flow[int_index, start_slice:dst_slice])
        )
        rho[int_index] = F_rain_flow[int_index] / (
            np.sqrt(F_rain[int_index]) * np.sqrt(F_flow[int_index])
        )
    pos_min = np.argmin(rho)
    Tr = pos_min + 1
    tol_fluct_rain = (rain_min / (2 * Tr + 1)) * Tr
    tol_fluct_flow = flow_int[-1] / 1e15
    fluct_rain[pos_min, np.fabs(fluct_rain[pos_min, :]) < tol_fluct_rain] = 0
    fluct_flow[pos_min, np.fabs(fluct_flow[pos_min, :]) < tol_fluct_flow] = 0
    fluct_rain_Tr = fluct_rain[pos_min, :]
    fluct_flow_Tr = fluct_flow[pos_min, :]
    fluct_bivariate_Tr = fluct_rain_Tr * fluct_flow_Tr
    fluct_bivariate_Tr[np.fabs(fluct_bivariate_Tr) < np.finfo(np.float64).eps] = (
        0  # 便于比较
    )
    return Tr, fluct_rain_Tr, fluct_flow_Tr, fluct_bivariate_Tr


def step3_core_identification(fluct_bivariate_Tr):
    d = np.diff(
        fluct_bivariate_Tr, prepend=[0], append=[0]
    )  # 计算相邻数值差分，为0代表两端点处于0区间
    d[np.fabs(d) < np.finfo(np.float64).eps] = 0  # 确保计算正确
    d = np.logical_not(d)  # 求0-1数组，为真代表为0区间
    d0 = np.logical_not(
        np.convolve(d, [1, 1], "valid")
    )  # 对相邻元素做OR，代表原数组数值是否处于某一0区间，再取反表示取有效值
    valid = np.logical_or(fluct_bivariate_Tr, d0)  # 有效core
    d_ = np.diff(valid, prepend=[0], append=[0])  # 求差分方便取上下边沿
    beginning_core = np.argwhere(d_ == 1)  # 上边沿为begin
    end_core = np.argwhere(d_ == -1) - 1  # 下边沿为end
    return beginning_core, end_core


def step4_end_rain_events(beginning_core, end_core, rain, fluct_rain_Tr, rain_min):
    end_rain = end_core.copy()
    rain = rain.T
    for g in range(end_core.size):
        if end_core[g] + 2 < fluct_rain_Tr.size and (
            np.fabs(fluct_rain_Tr[end_core[g] + 1]) < np.finfo(np.float64).eps
            and np.fabs(fluct_rain_Tr[end_core[g] + 2]) < np.finfo(np.float64).eps
        ):
            # case 1&2
            if np.fabs(rain[end_core[g]]) < np.finfo(np.float64).eps:
                # case 1
                while (
                    end_rain[g] > beginning_core[g]
                    and np.fabs(rain[end_rain[g]]) < np.finfo(np.float64).eps
                ):
                    end_rain[g] = end_rain[g] - 1
            else:
                # case 2
                bound = (
                    beginning_core[g + 1] if g + 1 < beginning_core.size else rain.size
                )
                while end_rain[g] < bound and rain[end_rain[g]] > rain_min:
                    end_rain[g] = end_rain[g] + 1
                end_rain[g] = end_rain[g] - 1  # 回到最后一个
        else:
            # case 3
            # 若在降水，先跳过
            while end_rain[g] >= beginning_core[g] and rain[end_rain[g]] > rain_min:
                end_rain[g] = end_rain[g] - 1
            while end_rain[g] >= beginning_core[g] and rain[end_rain[g]] < rain_min:
                end_rain[g] = end_rain[g] - 1
    return end_rain


def step5_beginning_rain_events(
    beginning_core, end_rain, rain, fluct_rain_Tr, rain_min
):
    beginning_rain = beginning_core.copy()
    rain = rain.T
    for g in range(beginning_core.size):
        if (
            beginning_core[g] >= 2
            and np.fabs(fluct_rain_Tr[beginning_core[g] - 1]) < np.finfo(np.float64).eps
            and np.fabs(fluct_rain_Tr[beginning_core[g] - 2]) < np.finfo(np.float64).eps
            and np.fabs(rain[beginning_core[g]]) < np.finfo(np.float64).eps
        ):
            # case 1
            while (
                beginning_rain[g] < end_rain[g]
                and np.fabs(rain[beginning_rain[g]]) < np.finfo(np.float64).eps
            ):
                beginning_rain[g] = beginning_rain[g] + 1
        else:
            # case 2&3
            bound = end_rain[g - 1] if g >= 1 else -1
            while beginning_rain[g] > bound and rain[beginning_rain[g]] > rain_min:
                beginning_rain[g] = beginning_rain[g] - 1
            beginning_rain[g] = beginning_rain[g] + 1  # 回到第一个
    return beginning_rain


def step6_checks_on_rain_events(
    beginning_rain, end_rain, rain, rain_min, beginning_core, end_core
):
    rain = rain.T
    beginning_rain = beginning_rain.copy()
    end_rain = end_rain.copy()
    if beginning_rain[0] == 0:  # 掐头
        beginning_rain = beginning_rain[1:]
        end_rain = end_rain[1:]
        beginning_core = beginning_core[1:]
        end_core = end_core[1:]
    if end_rain[-1] == rain.size - 1:  # 去尾
        beginning_rain = beginning_rain[:-2]
        end_rain = end_rain[:-2]
        beginning_core = beginning_core[:-2]
        end_core = end_core[:-2]
    error_time_reversed = beginning_rain > end_rain
    error_wrong_delimiter = np.logical_or(
        rain[beginning_rain - 1] > rain_min, rain[end_rain + 1] > rain_min
    )
    beginning_rain[error_time_reversed] = -2
    beginning_rain[error_wrong_delimiter] = -2
    end_rain[error_time_reversed] = -2
    end_rain[error_wrong_delimiter] = -2
    beginning_core[error_time_reversed] = -2
    beginning_core[error_wrong_delimiter] = -2
    end_core[error_time_reversed] = -2
    end_core[error_wrong_delimiter] = -2
    beginning_rain = beginning_rain[beginning_rain != -2]
    end_rain = end_rain[end_rain != -2]
    beginning_core = beginning_core[beginning_core != -2]
    end_core = end_core[end_core != -2]
    return beginning_rain, end_rain, beginning_core, end_core


def step7_end_flow_events(
    end_rain_checked, beginning_core, end_core, rain, fluct_rain_Tr, fluct_flow_Tr, Tr
):
    end_flow = np.empty(end_core.size, dtype=int)
    for g in range(end_rain_checked.size):
        if end_core[g] + 2 < fluct_rain_Tr.size and (
            np.fabs(fluct_rain_Tr[end_core[g] + 1]) < np.finfo(np.float64).eps
            and np.fabs(fluct_rain_Tr[end_core[g] + 2]) < np.finfo(np.float64).eps
        ):
            # case 1
            end_flow[g] = end_rain_checked[g]
            bound = (
                beginning_core[g + 1] + Tr if g + 1 < beginning_core.size else rain.size
            )
            bound = min(bound, rain.size)  # 防溢出
            # 若flow为负，先跳过
            while end_flow[g] < bound and fluct_flow_Tr[end_flow[g]] <= 0:
                end_flow[g] = end_flow[g] + 1
            while end_flow[g] < bound and fluct_flow_Tr[end_flow[g]] > 0:
                end_flow[g] = end_flow[g] + 1
            end_flow[g] = end_flow[g] - 1  # 回到最后一个
        else:
            # case 2
            end_flow[g] = end_core[g]
            while end_flow[g] >= beginning_core[g] and fluct_flow_Tr[end_flow[g]] <= 0:
                end_flow[g] = end_flow[g] - 1
    return end_flow


def step8_beginning_flow_events(
    beginning_rain_checked,
    end_rain_checked,
    rain,
    beginning_core,
    fluct_rain_Tr,
    fluct_flow_Tr,
):
    beginning_flow = np.empty(beginning_rain_checked.size, dtype=int)
    for g in range(beginning_rain_checked.size):
        if beginning_core[g] >= 2 and (
            np.fabs(fluct_rain_Tr[beginning_core[g] - 1]) < np.finfo(np.float64).eps
            and np.fabs(fluct_rain_Tr[beginning_core[g] - 2]) < np.finfo(np.float64).eps
        ):
            beginning_flow[g] = beginning_rain_checked[g]  # case 1
        else:
            beginning_flow[g] = beginning_core[g]  # case 2
        while (
            beginning_flow[g] < end_rain_checked[g]
            and fluct_flow_Tr[beginning_flow[g]] >= 0
        ):
            beginning_flow[g] = beginning_flow[g] + 1
    return beginning_flow


def step9_checks_on_flow_events(
    beginning_rain_checked, end_rain_checked, beginning_flow, end_flow, fluct_flow_Tr
):
    beginning_flow_checked = np.empty_like(beginning_flow, dtype=object)
    end_flow_checked = np.empty_like(end_flow, dtype=object)
    for g in range(len(beginning_flow)):
        if (
            not np.isnan(beginning_flow[g])
            and not np.isnan(end_flow[g])
            and (
                end_flow[g] <= beginning_flow[g]
                or fluct_flow_Tr[beginning_flow[g]] > 0
                or fluct_flow_Tr[end_flow[g]] < 0
                or beginning_flow[g] < beginning_rain_checked[g]
                or end_flow[g] < end_rain_checked[g]
            )
        ):
            beginning_flow_checked[g] = np.nan
            end_flow_checked[g] = np.nan
        else:
            beginning_flow_checked[g] = beginning_flow[g]
            end_flow_checked[g] = end_flow[g]

    index_events = [
        i
        for i in range(len(beginning_rain_checked))
        if not np.isnan(beginning_rain_checked[i])
        and not np.isnan(beginning_flow_checked[i])
        and not np.isnan(end_rain_checked[i])
        and not np.isnan(end_flow_checked[i])
    ]

    beginning_flow_ungrouped = beginning_flow_checked[index_events]
    end_flow_ungrouped = end_flow_checked[index_events]
    beginning_rain_ungrouped = beginning_rain_checked[index_events]
    end_rain_ungrouped = end_rain_checked[index_events]
    return (
        beginning_rain_ungrouped,
        end_rain_ungrouped,
        beginning_flow_ungrouped,
        end_flow_ungrouped,
    )
    # The following code is written by Wang, Yang
    # error_time_reversed = beginning_flow > end_flow
    # error_wrong_fluct = np.logical_or(
    #     np.logical_or(fluct_flow_Tr[beginning_flow] > 0, fluct_flow_Tr[end_flow] < 0),
    #     np.logical_or(
    #         beginning_flow < beginning_rain_checked, end_flow < end_rain_checked
    #     ),
    # )
    # beginning_flow[error_time_reversed] = -3
    # beginning_flow[error_wrong_fluct] = -3
    # end_flow[error_time_reversed] = -3
    # end_flow[error_wrong_fluct] = -3
    # beginning_flow = beginning_flow[beginning_flow != -3]
    # end_flow = end_flow[end_flow != -3]

    # beginning_rain_checked[error_time_reversed] = -3
    # beginning_rain_checked[error_wrong_fluct] = -3
    # end_rain_checked[error_time_reversed] = -3
    # end_rain_checked[error_wrong_fluct] = -3
    # beginning_rain = beginning_rain_checked[beginning_rain_checked != -3]
    # end_rain = end_rain_checked[end_rain_checked != -3]
    # return beginning_rain, end_rain, beginning_flow, end_flow


def step10_checks_on_overlapping_events(
    beginning_rain_ungrouped,
    end_rain_ungrouped,
    beginning_flow_ungrouped,
    end_flow_ungrouped,
    time,
):
    # rain
    # order1 = np.reshape(
    #     np.hstack(
    #         (
    #             np.reshape(beginning_rain_ungrouped, (-1, 1)),
    #             np.reshape(end_rain_ungrouped, (-1, 1)),
    #         )
    #     ),
    #     (1, -1),
    # )
    # reversed1 = np.diff(order1) <= 0
    # order1[np.hstack((reversed1, [[False]]))] = -2
    # order1[np.hstack(([[False]], reversed1))] = -2
    # order1 = order1[order1 != -2]
    # # flow
    # order2 = np.reshape(
    #     np.hstack(
    #         (
    #             np.reshape(beginning_flow_ungrouped, (-1, 1)),
    #             np.reshape(end_flow_ungrouped, (-1, 1)),
    #         )
    #     ),
    #     (1, -1),
    # )
    # reversed2 = np.diff(order2) <= 0
    # order2[np.hstack((reversed2, [[False]]))] = -3
    # order2[np.hstack(([[False]], reversed2))] = -3
    # order2 = order2[order2 != -3]
    # # group
    # rain_grouped = np.reshape(order1, (-1, 2)).T
    # beginning_rain_grouped = rain_grouped[0]
    # end_rain_grouped = rain_grouped[1]
    # flow_grouped = np.reshape(order2, (-1, 2)).T
    # beginning_flow_grouped = flow_grouped[0]
    # end_flow_grouped = flow_grouped[1]
    # return (
    #     time[beginning_rain_grouped],
    #     time[end_rain_grouped],
    #     time[beginning_flow_grouped],
    #     time[end_flow_grouped],
    # )

    beginning_rain_ungrouped = beginning_rain_ungrouped.astype(float)
    beginning_flow_ungrouped = beginning_flow_ungrouped.astype(float)
    end_rain_ungrouped = end_rain_ungrouped.astype(float)
    end_flow_ungrouped = end_flow_ungrouped.astype(float)
    q = 1
    marker_overlapping = []
    for g in range(len(end_rain_ungrouped) - 1):
        if (
            end_rain_ungrouped[g] > beginning_rain_ungrouped[g + 1]
            or end_flow_ungrouped[g] > beginning_flow_ungrouped[g + 1]
        ):
            marker_overlapping.append(g)
            q += 1
    if marker_overlapping:
        q = 0
        while q < len(marker_overlapping) - 1:
            to_group = [marker_overlapping[q]]
            while (
                q < len(marker_overlapping)
                and marker_overlapping[q] == marker_overlapping[q + 1] - 1
            ):
                to_group.append(marker_overlapping[q + 1])
                q += 1

            beginning_rain_ungrouped[to_group[0]] = beginning_rain_ungrouped[
                to_group[0]
            ]
            beginning_flow_ungrouped[to_group[0]] = beginning_flow_ungrouped[
                to_group[0]
            ]
            end_flow_ungrouped[to_group[0]] = end_flow_ungrouped[to_group[-1] + 1]
            end_rain_ungrouped[to_group[0]] = end_rain_ungrouped[to_group[-1] + 1]

            if len(to_group) > 1:
                beginning_rain_ungrouped[to_group[1:]] = np.nan
                beginning_flow_ungrouped[to_group[1:]] = np.nan
                end_flow_ungrouped[to_group[1:]] = np.nan
                end_rain_ungrouped[to_group[1:]] = np.nan

            beginning_rain_ungrouped[to_group[-1] + 1] = np.nan
            beginning_flow_ungrouped[to_group[-1] + 1] = np.nan
            end_flow_ungrouped[to_group[-1] + 1] = np.nan
            end_rain_ungrouped[to_group[-1] + 1] = np.nan

            to_group = []
            q += 1

    index_events2 = np.where(
        ~np.isnan(beginning_rain_ungrouped)
        & ~np.isnan(beginning_flow_ungrouped)
        & ~np.isnan(end_rain_ungrouped)
        & ~np.isnan(end_flow_ungrouped)
    )[0]
    beginning_flow_grouped = beginning_flow_ungrouped[index_events2]
    end_flow_grouped = end_flow_ungrouped[index_events2]
    beginning_rain_grouped = beginning_rain_ungrouped[index_events2]
    end_rain_grouped = end_rain_ungrouped[index_events2]
    beginning_rain_grouped = beginning_rain_grouped.astype(int)
    end_rain_grouped = end_rain_grouped.astype(int)
    beginning_flow_grouped = beginning_flow_grouped.astype(int)
    end_flow_grouped = end_flow_grouped.astype(int)
    beginning_rain = time[beginning_rain_grouped]
    end_rain = time[end_rain_grouped]
    end_flow = time[end_flow_grouped]
    beginning_flow = time[beginning_flow_grouped]
    return beginning_rain, end_rain, beginning_flow, end_flow
    # biliu_rain_division = pd.DataFrame({'BEGINNING_RAIN': BEGINNING_RAIN, 'END_RAIN': END_RAIN})
    # biliu_rain_division.to_csv(os.path.join(definitions.ROOT_DIR, 'example/yingnariver_data/yingna_rain_flow_division/yingna_rain_division.txt'), sep='\t')
    # biliu_flow_division = pd.DataFrame({'BEGINNING_FLOW': BEGINNING_FLOW, 'END_FLOW': END_FLOW})
    # biliu_flow_division.to_csv(os.path.join(definitions.ROOT_DIR, 'example/yingnariver_data/yingna_rain_flow_division/yingna_flow_division.txt'), sep='\t')


def baseflow_curve(beginning_flow, end_flow, flow, time):
    baseflow = np.copy(flow)
    beg_end_series = np.array([], dtype=beginning_flow[0].dtype)
    for j in range(len(beginning_flow)):
        beg_end_series = np.concatenate(
            (beg_end_series, [beginning_flow[j], end_flow[j]])
        )

    for k in range(len(beg_end_series) - 1):
        index_beg = np.where(time == beg_end_series[k])[0]
        index_end = np.where(time == beg_end_series[k + 1])[0]
        index_beg = int(index_beg)
        index_end = int(index_end)
        if (
            len(np.where(np.isnan(flow[index_beg : index_end + 1]) == 1)[0])
            >= len(flow[index_beg : index_end + 1]) * 0.9
        ):
            baseflow[index_beg : index_end + 1] = np.nan
        elif index_end - index_beg == 1:
            baseflow[index_beg] = flow[index_beg]
            baseflow[index_end] = flow[index_end]
        elif flow[index_beg] < flow[index_end]:
            increment = (flow[index_end] - flow[index_beg]) / (index_end - index_beg)
            for m in range(index_beg + 1, index_end):
                baseflow[m] = baseflow[index_beg] + increment * (m - index_beg)
        elif flow[index_beg] > flow[index_end]:
            increment = (flow[index_beg] - flow[index_end]) / (index_end - index_beg)
            for m in range(index_beg + 1, index_end):
                baseflow[m] = baseflow[index_beg] - increment * (m - index_beg)

    for m in range(len(baseflow)):
        baseflow[m] = min(baseflow[m], flow[m])
    return baseflow


def step11_event_analysis(
    beginning_rain,
    end_rain,
    beginning_flow,
    end_flow,
    rain,
    flow,
    time,
    flow_threshold,
    multiple=1,
    flag=0,
    duration_max=2400,
):
    """The step 11 is a new step for the DMCA-ESR method
    we proposed to check again for the rainfall-runoff events.

    Parameters
    ----------
    beginning_rain : _type_
        _description_
    end_rain : _type_
        _description_
    beginning_flow : _type_
        _description_
    end_flow : _type_
        _description_
    rain : _type_
        _description_
    flow : _type_
        _description_
    time : _type_
        _description_
    flag : int, by default 0
        if flag != 1, calculation target is flow - baseflow,
        if flag == 1, cal target is flow
        the default value 0 is recommended as flag == 1 is just for comparison
    multiple : int
        1h means the multiple is 1, 1d means the multiple is 24
    flow_threshold : float
        flow threshold -- remove events with max flow lower than it
        its unit is mm_h
    duration_max : int
        maximum duration of the rainfall-runoff events we set; unit is hour

    Returns
    -------
    _type_
        _description_
    """
    duration_rain = np.zeros(len(beginning_rain))
    duration_runoff = np.zeros(len(beginning_flow))
    volume_rain = np.zeros(len(beginning_rain))
    volume_runoff = np.zeros(len(beginning_flow))
    runoff_ratio = np.zeros(len(beginning_rain))
    for h in range(len(beginning_rain)):
        duration_rain[h] = (
            (np.datetime64(end_rain[h]) - np.datetime64(beginning_rain[h]))
            / np.timedelta64(1, "s")
            / (60 * 60 * multiple)
        )
        duration_runoff[h] = (
            (np.datetime64(end_flow[h]) - np.datetime64(beginning_flow[h]))
            / np.timedelta64(1, "s")
            / (60 * 60 * multiple)
        )
        index_beginning_event = np.where(time == beginning_rain[h])[0]
        index_end_event = np.where(time == end_rain[h])[0]
        index_beginning_event = int(index_beginning_event)
        index_end_event = int(index_end_event)
        volume_rain[h] = (
            np.nansum(rain[index_beginning_event:index_end_event]) * multiple
        )

    if flag == 1:
        for h in range(len(beginning_flow)):
            index_beginning_event = np.where(time == beginning_flow[h])[0]
            index_end_event = np.where(time == end_flow[h])[0]
            volume_runoff[h] = (
                np.nansum(flow[index_beginning_event:index_end_event]) * multiple
            )
    else:
        baseflow = baseflow_curve(beginning_flow, end_flow, flow, time)
        for h in range(len(beginning_flow)):
            index_beginning_event = np.where(time == beginning_flow[h])[0]
            index_end_event = np.where(time == end_flow[h])[0]
            index_beginning_event = int(index_beginning_event)
            index_end_event = int(index_end_event)

            q = flow[index_beginning_event:index_end_event]
            qb = baseflow[index_beginning_event:index_end_event]
            volume_runoff[h] = np.nansum(q - qb) * multiple
    runoff_ratio = volume_runoff / volume_rain
    result_df = pd.DataFrame(
        {
            "BEGINNING_RAIN": beginning_rain,
            "END_RAIN": end_rain,
            "BEGINNING_FLOW": beginning_flow,
            "END_FLOW": end_flow,
            "DURATION_RAIN": duration_rain,
            "DURATION_RUNOFF": duration_runoff,
            "VOLUME_RAIN": volume_rain,
            "VOLUME_RUNOFF": volume_runoff,
            "RUNOFF_RATIO": runoff_ratio,
        }
    )
    drop_list = []
    for i in range(len(beginning_rain)):
        index_beginning_event = np.where(time == beginning_flow[i])[0]
        index_end_event = np.where(time == end_flow[i])[0]
        index_beginning_event = int(index_beginning_event)
        index_end_event = int(index_end_event)
        if flow[index_beginning_event:index_end_event].max() < flow_threshold:
            drop_list.append(i)
        if result_df["DURATION_RAIN"].iloc[i] > duration_max:
            drop_list.append(i)
        if result_df["DURATION_RUNOFF"].iloc[i] > duration_max:
            drop_list.append(i)
        if result_df["VOLUME_RAIN"].iloc[i] == 0:
            drop_list.append(i)
        if result_df["VOLUME_RUNOFF"].iloc[i] == 0:
            drop_list.append(i)
        if runoff_ratio[i] > 1:
            drop_list.append(i)
    drop_array = np.unique(np.array(drop_list, dtype=int))
    result_df = result_df.drop(index=drop_array)
    duration_rain = result_df["DURATION_RAIN"]
    volume_rain = result_df["VOLUME_RAIN"]
    duration_runoff = result_df["DURATION_RUNOFF"]
    volume_runoff = result_df["VOLUME_RUNOFF"]
    runoff_ratio = result_df["RUNOFF_RATIO"]
    beginning_rain = result_df["BEGINNING_RAIN"]
    end_rain = result_df["END_RAIN"]
    beginning_flow = result_df["BEGINNING_FLOW"]
    end_flow = result_df["END_FLOW"]
    return (
        duration_rain,
        volume_rain,
        duration_runoff,
        volume_runoff,
        runoff_ratio,
        beginning_rain,
        end_rain,
        beginning_flow,
        end_flow,
    )


def rainfall_runoff_event_identify(
    rain, flow, rain_min=0.02, max_window=100, multiple=24, flow_threshold=0.02
):
    """Full process for identification of rainfall-runoff events
    We make function for each step, name the function as the step number and
    add some anther necessary steps for the DMCA-ESR method.

    Parameters
    ----------
    rain : pd.Series
        rainfall time series, for the original data,
        unit must be same as the time interval.
        For example, when unit is mm/day, time interval is day
    flow : pd.Series
        streamflow time series,
        the requirement for unit is same as the rainfall
    rain_min : float
        minimum rainfall threshold; its unit is mm/h
    max_window : int
        maximum window size in the first two steps
        it has no unit, just meaning the number of time intervals,
        no matter the time unit is hour, day, or others
    multiple : int
        24 for daily data, meaning convert unit from mm/day to mm/h
        1 for hourly data
    flow_threshold : float
        events with max flow lower than it will be removed; its unit is mm/h
    """
    time = rain.index.to_numpy()
    rain = rain.to_numpy() / multiple
    flow = flow.to_numpy() / multiple
    (
        Tr,
        fluct_rain_Tr,
        fluct_flow_Tr,
        fluct_bivariate_Tr,
    ) = step1_step2_tr_and_fluctuations_timeseries(rain, flow, rain_min, max_window)
    beginning_core, end_core = step3_core_identification(fluct_bivariate_Tr)
    end_rain = step4_end_rain_events(
        beginning_core, end_core, rain, fluct_rain_Tr, rain_min
    )
    beginning_rain = step5_beginning_rain_events(
        beginning_core, end_rain, rain, fluct_rain_Tr, rain_min
    )
    (
        beginning_rain_checked,
        end_rain_checked,
        beginning_core,
        end_core,
    ) = step6_checks_on_rain_events(
        beginning_rain, end_rain, rain, rain_min, beginning_core, end_core
    )
    end_flow = step7_end_flow_events(
        end_rain_checked,
        beginning_core,
        end_core,
        rain,
        fluct_rain_Tr,
        fluct_flow_Tr,
        Tr,
    )
    beginning_flow = step8_beginning_flow_events(
        beginning_rain_checked,
        end_rain_checked,
        rain,
        beginning_core,
        fluct_rain_Tr,
        fluct_flow_Tr,
    )
    (
        beginning_rain_ungrouped,
        end_rain_ungrouped,
        beginning_flow_ungrouped,
        end_flow_ungrouped,
    ) = step9_checks_on_flow_events(
        beginning_rain_checked,
        end_rain_checked,
        beginning_flow,
        end_flow,
        fluct_flow_Tr,
    )
    (
        beginning_rain,
        end_rain,
        beginning_flow,
        end_flow,
    ) = step10_checks_on_overlapping_events(
        beginning_rain_ungrouped,
        end_rain_ungrouped,
        beginning_flow_ungrouped,
        end_flow_ungrouped,
        time,
    )
    (
        duration_rain,
        volume_rain,
        duration_runoff,
        volume_runoff,
        runoff_ratio,
        beginning_rain,
        end_rain,
        beginning_flow,
        end_flow,
    ) = step11_event_analysis(
        beginning_rain,
        end_rain,
        beginning_flow,
        end_flow,
        rain,
        flow,
        time,
        flow_threshold,
        multiple=multiple,
    )
    print(beginning_rain, end_rain, beginning_flow, end_flow)
    print(len(beginning_rain), len(end_rain), len(beginning_flow), len(end_flow))
    return pd.DataFrame(
        {
            "BEGINNING_RAIN": beginning_rain,
            "END_RAIN": end_rain,
            "DURATION_RAIN": duration_rain,
            "BEGINNING_FLOW": beginning_flow,
            "END_FLOW": end_flow,
            "DURATION_RUNOFF": duration_runoff,
            "VOLUME_RAIN": volume_rain,
            "VULUME_RUNOFF": volume_runoff,
            "RUNOFF_RATIO": runoff_ratio,
        }
    )



"""
Author: Wenyu Ouyang
Date: 2024-08-09 19:38:52
LastEditTime: 2024-09-15 14:51:07
LastEditors: Wenyu Ouyang
Description: Some constants for data source
FilePath: \hydrodatasource\hydrodatasource\configs\data_consts.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

ERA5LAND_VARS = [
    "dewpoint_temperature_2m",
    "temperature_2m",
    "skin_temperature",
    "soil_temperature_level_1",
    "soil_temperature_level_2",
    "soil_temperature_level_3",
    "soil_temperature_level_4",
    "lake_bottom_temperature",
    "lake_ice_depth",
    "lake_ice_temperature",
    "lake_mix_layer_depth",
    "lake_mix_layer_temperature",
    "lake_shape_factor",
    "lake_total_layer_temperature",
    "snow_albedo",
    "snow_cover",
    "snow_density",
    "snow_depth_hourly",
    "snow_depth_water_equivalent",
    "snowfall",
    "snowmelt",
    "temperature_of_snow_layer",
    "skin_reservoir_content",
    "volumetric_soil_water_layer_1",
    "volumetric_soil_water_layer_2",
    "volumetric_soil_water_layer_3",
    "volumetric_soil_water_layer_4",
    "forecast_albedo",
    "surface_latent_heat_flux",
    "surface_net_solar_radiation",
    "surface_net_thermal_radiation",
    "surface_sensible_heat_flux",
    "surface_solar_radiation_downwards",
    "surface_thermal_radiation_downwards",
    "evaporation_from_bare_soil",
    "evaporation_from_open_water_surfaces_excluding_oceans",
    "evaporation_from_the_top_of_canopy",
    "evaporation_from_vegetation_transpiration",
    "potential_evaporation",
    "runoff",
    "snow_evaporation",
    "sub_surface_runoff",
    "surface_runoff",
    "total_evaporation",
    "u_component_of_wind_10m",
    "v_component_of_wind_10m",
    "surface_pressure",
    "total_precipitation",
    "leaf_area_index_high_vegetation",
    "leaf_area_index_low_vegetation",
    "snowfall_hourly",
    "snowmelt_hourly",
    "surface_latent_heat_flux_hourly",
    "surface_net_solar_radiation_hourly",
    "surface_net_thermal_radiation_hourly",
    "surface_sensible_heat_flux_hourly",
    "surface_solar_radiation_downwards_hourly",
    "surface_thermal_radiation_downwards_hourly",
    "evaporation_from_bare_soil_hourly",
    "evaporation_from_open_water_surfaces_excluding_oceans_hourly",
    "evaporation_from_the_top_of_canopy_hourly",
    "evaporation_from_vegetation_transpiration_hourly",
    "potential_evaporation_hourly",
    "runoff_hourly",
    "snow_evaporation_hourly",
    "sub_surface_runoff_hourly",
    "surface_runoff_hourly",
    "total_evaporation_hourly",
    "total_precipitation_hourly",
]

ERA5LAND_ET_REALATED_VARS = [
    "evaporation_from_bare_soil",
    "evaporation_from_open_water_surfaces_excluding_oceans",
    "evaporation_from_the_top_of_canopy",
    "evaporation_from_vegetation_transpiration",
    "potential_evaporation",
    "snow_evaporation",
    "total_evaporation",
    "evaporation_from_bare_soil_hourly",
    "evaporation_from_open_water_surfaces_excluding_oceans_hourly",
    "evaporation_from_the_top_of_canopy_hourly",
    "evaporation_from_vegetation_transpiration_hourly",
    "potential_evaporation_hourly",
    "snow_evaporation_hourly",
    "total_evaporation_hourly",
]

MODIS_ET_PET_8D_VARS = [
    "ET_modis16a2006",
    "ET_modis16a2gf061",
    "PET_modis16a2006",
    "PET_modis16a2gf061",
]
